from abc import abstractmethod
import scipy.special
import json
import struct
import cext_acv
import warnings
from .py_acv import *
from .utils import rebuild_tree
from distutils.version import LooseVersion
from sklearn.utils.validation import check_array


# This is based on https://github.com/slundberg/shap/blob/master/shap/explainers/_tree.py
class BaseTree:
    """ An ensemble of decision trees.

    This object provides a common interface to many different types of models.
    """

    def __init__(self, model, data=None, data_missing=None, cache=False, cache_normalized=False, multi_threads=True,
                 C=[[]]):
        self.model_type = "internal"
        self.trees = None
        self.base_offset = 0
        self.model_output = None
        self.objective = None  # what we explain when explaining the loss of the model
        self.tree_output = None  # what are the units of the values in the leaves of the trees
        self.internal_dtype = np.float64
        self.input_dtype = np.float64  # for sklearn we need to use np.float32 to always get exact matches to their predictions
        self.data = check_array(data, dtype=[np.double])
        self.data_missing = data_missing
        self.fully_defined_weighting = True  # does the background dataset land in every leaf (making it valid for the tree_path_dependent method)
        self.tree_limit = None  # used for limiting the number of trees we use by default (like from early stopping)
        self.num_stacked_models = 1  # If this is greater than 1 it means we have multiple stacked models with the same number of trees in each model (XGBoost multi-output style)
        self.cat_feature_indices = None  # If this is set it tells us which features are treated categorically
        self.model = model
        self.cache = cache
        self.cache_normalized = cache_normalized
        self.C = C
        self.multi_threads = multi_threads
        # we use names like keras
        objective_name_map = {
            "mse": "squared_error",
            "variance": "squared_error",
            "friedman_mse": "squared_error",
            "reg:linear": "squared_error",
            "reg:squarederror": "squared_error",
            "regression": "squared_error",
            "regression_l2": "squared_error",
            "mae": "absolute_error",
            "gini": "binary_crossentropy",
            "entropy": "binary_crossentropy",
            "reg:logistic": "binary_crossentropy",
            "binary:logistic": "binary_crossentropy",
            "binary_logloss": "binary_crossentropy",
            "binary": "binary_crossentropy"
        }

        tree_output_name_map = {
            "regression": "raw_value",
            "regression_l2": "squared_error",
            "reg:linear": "raw_value",
            "reg:squarederror": "raw_value",
            "reg:logistic": "log_odds",
            "binary:logistic": "log_odds",
            "binary_logloss": "log_odds",
            "binary": "log_odds"
        }

        if type(model) is dict and "trees" in model:
            # This allows a dictionary to be passed that represents the model.
            # this dictionary has several numerica paramters and also a list of trees
            # where each tree is a dictionary describing that tree
            if "internal_dtype" in model:
                self.internal_dtype = model["internal_dtype"]
            if "input_dtype" in model:
                self.input_dtype = model["input_dtype"]
            if "objective" in model:
                self.objective = model["objective"]
            if "tree_output" in model:
                self.tree_output = model["tree_output"]
            if "base_offset" in model:
                self.base_offset = model["base_offset"]
            self.trees = [SingleTree(t, data=data, data_missing=data_missing) for t in model["trees"]]
        elif type(model) is list and type(model[0]) == SingleTree:  # old-style direct-load format
            self.trees = model
        elif safe_isinstance(model,
                             ["sklearn.ensemble.RandomForestRegressor", "sklearn.ensemble.forest.RandomForestRegressor",
                              "econml.grf._base_grf.BaseGRF"]):
            assert hasattr(model, "estimators_"), "Model has no `estimators_`! Have you called `model.fit`?"
            self.internal_dtype = model.estimators_[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [SingleTree(e.tree_, scaling=scaling, data=data, data_missing=data_missing) for e in
                          model.estimators_]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "raw_value"
        elif safe_isinstance(model, ["sklearn.ensemble.IsolationForest", "sklearn.ensemble._iforest.IsolationForest"]):
            self.dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [IsoTree(e.tree_, f, scaling=scaling, data=data, data_missing=data_missing) for e, f in
                          zip(model.estimators_, model.estimators_features_)]
            self.tree_output = "raw_value"
        elif safe_isinstance(model, ["pyod.models.iforest.IForest"]):
            self.dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [IsoTree(e.tree_, f, scaling=scaling, data=data, data_missing=data_missing) for e, f in
                          zip(model.detector_.estimators_, model.detector_.estimators_features_)]
            self.tree_output = "raw_value"
        elif safe_isinstance(model, "skopt.learning.forest.RandomForestRegressor"):
            assert hasattr(model, "estimators_"), "Model has no `estimators_`! Have you called `model.fit`?"
            self.internal_dtype = model.estimators_[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [SingleTree(e.tree_, scaling=scaling, data=data, data_missing=data_missing) for e in
                          model.estimators_]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "raw_value"
        elif safe_isinstance(model,
                             ["sklearn.ensemble.ExtraTreesRegressor", "sklearn.ensemble.forest.ExtraTreesRegressor"]):
            assert hasattr(model, "estimators_"), "Model has no `estimators_`! Have you called `model.fit`?"
            self.internal_dtype = model.estimators_[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [SingleTree(e.tree_, scaling=scaling, data=data, data_missing=data_missing) for e in
                          model.estimators_]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "raw_value"
        elif safe_isinstance(model, "skopt.learning.forest.ExtraTreesRegressor"):
            assert hasattr(model, "estimators_"), "Model has no `estimators_`! Have you called `model.fit`?"
            self.internal_dtype = model.estimators_[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [SingleTree(e.tree_, scaling=scaling, data=data, data_missing=data_missing) for e in
                          model.estimators_]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "raw_value"
        elif safe_isinstance(model, ["sklearn.tree.DecisionTreeRegressor", "sklearn.tree.tree.DecisionTreeRegressor",
                                     "econml.grf._base_grftree.GRFTree"]):
            self.internal_dtype = model.tree_.value.dtype.type
            self.input_dtype = np.float32
            self.trees = [SingleTree(model.tree_, data=data, data_missing=data_missing)]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "raw_value"
            # self.scaling = 1
        elif safe_isinstance(model,
                             ["sklearn.tree.DecisionTreeClassifier", "sklearn.tree.tree.DecisionTreeClassifier"]):
            self.internal_dtype = model.tree_.value.dtype.type
            self.input_dtype = np.float32
            self.trees = [SingleTree(model.tree_, normalize=True, data=data, data_missing=data_missing)]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "probability"
            # self.scaling = 1
        elif safe_isinstance(model, ["sklearn.ensemble.RandomForestClassifier",
                                     "sklearn.ensemble.forest.RandomForestClassifier"]):
            assert hasattr(model, "estimators_"), "Model has no `estimators_`! Have you called `model.fit`?"
            self.internal_dtype = model.estimators_[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [SingleTree(e.tree_, normalize=True, scaling=scaling, data=data, data_missing=data_missing) for
                          e in model.estimators_]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "probability"
        elif safe_isinstance(model, ["skranger.ensemble.classifier.RangerForestClassifier"]):
            assert hasattr(model, "estimators_"), "Model has no `estimators_`! Have you called `model.fit`?"
            self.internal_dtype = model.estimators_[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [SingleTree(e.tree_, normalize=True, scaling=scaling, data=data, data_missing=data_missing) for
                          e in model.estimators_]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "probability"
        elif safe_isinstance(model, ["sklearn.ensemble.ExtraTreesClassifier",
                                     "sklearn.ensemble.forest.ExtraTreesClassifier"]):  # TODO: add unit test for this case
            assert hasattr(model, "estimators_"), "Model has no `estimators_`! Have you called `model.fit`?"
            self.internal_dtype = model.estimators_[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [SingleTree(e.tree_, normalize=True, scaling=scaling, data=data, data_missing=data_missing) for
                          e in model.estimators_]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "probability"
        elif safe_isinstance(model, ["sklearn.ensemble.GradientBoostingRegressor",
                                     "sklearn.ensemble.gradient_boosting.GradientBoostingRegressor"]):
            self.input_dtype = np.float32

            # currently we only support the mean and quantile estimators
            if safe_isinstance(model.init_,
                               ["sklearn.ensemble.MeanEstimator", "sklearn.ensemble.gradient_boosting.MeanEstimator"]):
                self.base_offset = model.init_.mean
            elif safe_isinstance(model.init_, ["sklearn.ensemble.QuantileEstimator",
                                               "sklearn.ensemble.gradient_boosting.QuantileEstimator"]):
                self.base_offset = model.init_.quantile
            elif safe_isinstance(model.init_, "sklearn.dummy.DummyRegressor"):
                self.base_offset = model.init_.constant_[0]
            else:
                assert False, "Unsupported init model type: " + str(type(model.init_))

            self.trees = [SingleTree(e.tree_, scaling=model.learning_rate, data=data, data_missing=data_missing) for e
                          in model.estimators_[:, 0]]
            # self.scaling = model.learning_rate
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "raw_value"
        elif safe_isinstance(model, ["sklearn.ensemble.HistGradientBoostingRegressor"]):
            import sklearn
            if self.model_output == "predict":
                self.model_output = "raw"
            self.input_dtype = sklearn.ensemble._hist_gradient_boosting.common.X_DTYPE
            self.base_offset = model._baseline_prediction
            self.trees = []
            for p in model._predictors:
                nodes = p[0].nodes
                # each node has values: ('value', 'count', 'feature_idx', 'threshold', 'missing_go_to_left', 'left', 'right', 'gain', 'depth', 'is_leaf', 'bin_threshold')
                tree = {
                    "children_left": np.array([-1 if n[9] else n[5] for n in nodes]),
                    "children_right": np.array([-1 if n[9] else n[6] for n in nodes]),
                    "children_default": np.array([-1 if n[9] else (n[5] if n[4] else n[6]) for n in nodes]),
                    "features": np.array([-2 if n[9] else n[2] for n in nodes]),
                    "thresholds": np.array([n[3] for n in nodes], dtype=np.float64),
                    "values": np.array([[n[0]] for n in nodes], dtype=np.float64),
                    "node_sample_weight": np.array([n[1] for n in nodes], dtype=np.float64),
                }
                self.trees.append(SingleTree(tree, data=data, data_missing=data_missing))
            self.objective = objective_name_map.get(model.loss, None)
            self.tree_output = "raw_value"
            # self.scaling = 1
        elif safe_isinstance(model, ["sklearn.ensemble.HistGradientBoostingClassifier"]):
            import sklearn
            self.base_offset = model._baseline_prediction
            if hasattr(self.base_offset, "__len__") and self.model_output != "raw":
                raise Exception(
                    "Multi-output HistGradientBoostingClassifier models are not yet supported unless model_output=\"raw\". See GitHub issue #1028")
            self.input_dtype = sklearn.ensemble._hist_gradient_boosting.common.X_DTYPE
            self.num_stacked_models = len(model._predictors[0])
            if self.model_output == "predict_proba":
                if self.num_stacked_models == 1:
                    self.model_output = "probability_doubled"  # with predict_proba we need to double the outputs to match
                else:
                    self.model_output = "probability"
            self.trees = []
            for p in model._predictors:
                for i in range(self.num_stacked_models):
                    nodes = p[i].nodes
                    # each node has values: ('value', 'count', 'feature_idx', 'threshold', 'missing_go_to_left', 'left', 'right', 'gain', 'depth', 'is_leaf', 'bin_threshold')
                    tree = {
                        "children_left": np.array([-1 if n[9] else n[5] for n in nodes]),
                        "children_right": np.array([-1 if n[9] else n[6] for n in nodes]),
                        "children_default": np.array([-1 if n[9] else (n[5] if n[4] else n[6]) for n in nodes]),
                        "features": np.array([-2 if n[9] else n[2] for n in nodes]),
                        "thresholds": np.array([n[3] for n in nodes], dtype=np.float64),
                        "values": np.array([[n[0]] for n in nodes], dtype=np.float64),
                        "node_sample_weight": np.array([n[1] for n in nodes], dtype=np.float64),
                    }
                    self.trees.append(SingleTree(tree, data=data, data_missing=data_missing))
            self.objective = objective_name_map.get(model.loss, None)
            self.tree_output = "log_odds"
            # self.scaling = 1
        elif safe_isinstance(model, ["sklearn.ensemble.GradientBoostingClassifier",
                                     "sklearn.ensemble._gb.GradientBoostingClassifier",
                                     "sklearn.ensemble.gradient_boosting.GradientBoostingClassifier"]):
            self.input_dtype = np.float32

            # TODO: deal with estimators for each class
            if model.estimators_.shape[1] > 1:
                assert False, "GradientBoostingClassifier is only supported for binary classification right now!"

            # currently we only support the logs odds estimator
            if safe_isinstance(model.init_, ["sklearn.ensemble.LogOddsEstimator",
                                             "sklearn.ensemble.gradient_boosting.LogOddsEstimator"]):
                self.base_offset = model.init_.prior
                self.tree_output = "log_odds"
            elif safe_isinstance(model.init_, "sklearn.dummy.DummyClassifier"):
                self.base_offset = scipy.special.logit(model.init_.class_prior_[
                                                           1])  # with two classes the trees only model the second class. # pylint: disable=no-member
                self.tree_output = "log_odds"
            else:
                assert False, "Unsupported init model type: " + str(type(model.init_))

            self.trees = [SingleTree(e.tree_, scaling=model.learning_rate, data=data, data_missing=data_missing) for e
                          in model.estimators_[:, 0]]
            # self.scaling = model.learning_rate
            self.objective = objective_name_map.get(model.criterion, None)
        elif "pyspark.ml" in str(type(model)):
            assert_import("pyspark")
            self.model_type = "pyspark"
            # model._java_obj.getImpurity() can be gini, entropy or variance.
            self.objective = objective_name_map.get(model._java_obj.getImpurity(), None)
            if "Classification" in str(type(model)):
                normalize = True
                self.tree_output = "probability"
            else:
                normalize = False
                self.tree_output = "raw_value"
            # Spark Random forest, create 1 weighted (avg) tree per sub-model
            if safe_isinstance(model, "pyspark.ml.classification.RandomForestClassificationModel") \
                    or safe_isinstance(model, "pyspark.ml.regression.RandomForestRegressionModel"):
                sum_weight = sum(model.treeWeights)  # output is average of trees
                self.trees = [SingleTree(tree, normalize=normalize, scaling=model.treeWeights[i] / sum_weight) for
                              i, tree in enumerate(model.trees)]
                # self.scaling = model.treeWeights[i] / sum_weight
                # Spark GBT, create 1 weighted (learning rate) tree per sub-model
            elif safe_isinstance(model, "pyspark.ml.classification.GBTClassificationModel") \
                    or safe_isinstance(model, "pyspark.ml.regression.GBTRegressionModel"):
                self.objective = "squared_error"  # GBT subtree use the variance
                self.tree_output = "raw_value"
                self.trees = [SingleTree(tree, normalize=False, scaling=model.treeWeights[i]) for i, tree in
                              enumerate(model.trees)]
            # Spark Basic model (single tree)
            elif safe_isinstance(model, "pyspark.ml.classification.DecisionTreeClassificationModel") \
                    or safe_isinstance(model, "pyspark.ml.regression.DecisionTreeRegressionModel"):
                self.trees = [SingleTree(model, normalize=normalize, scaling=1)]
            else:
                assert False, "Unsupported Spark model type: " + str(type(model))
        elif safe_isinstance(model, "xgboost.core.Booster"):
            import xgboost
            self.original_model = model
            self.model_type = "xgboost"
            xgb_loader = XGBTreeModelLoader(self.original_model)
            self.trees = xgb_loader.get_trees(data=data, data_missing=data_missing)
            self.base_offset = xgb_loader.base_score
            self.objective = objective_name_map.get(xgb_loader.name_obj, None)
            self.tree_output = tree_output_name_map.get(xgb_loader.name_obj, None)
            if xgb_loader.num_class > 0:
                self.num_stacked_models = xgb_loader.num_class
            # self.scaling = 1
        elif safe_isinstance(model, "xgboost.sklearn.XGBClassifier"):
            import xgboost
            self.input_dtype = np.float32
            self.model_type = "xgboost"
            self.original_model = model.get_booster()
            xgb_loader = XGBTreeModelLoader(self.original_model)
            self.trees = xgb_loader.get_trees(data=data, data_missing=data_missing)
            self.base_offset = xgb_loader.base_score
            self.objective = objective_name_map.get(xgb_loader.name_obj, None)
            self.tree_output = tree_output_name_map.get(xgb_loader.name_obj, None)
            self.tree_limit = getattr(model, "best_ntree_limit", None)
            if xgb_loader.num_class > 0:
                self.num_stacked_models = xgb_loader.num_class
            if self.model_output == "predict_proba":
                if self.num_stacked_models == 1:
                    self.model_output = "probability_doubled"  # with predict_proba we need to double the outputs to match
                else:
                    self.model_output = "probability"
            # self.scaling = 1
        elif safe_isinstance(model, "xgboost.sklearn.XGBRegressor"):
            import xgboost
            self.original_model = model.get_booster()
            self.model_type = "xgboost"
            xgb_loader = XGBTreeModelLoader(self.original_model)
            self.trees = xgb_loader.get_trees(data=data, data_missing=data_missing)
            self.base_offset = xgb_loader.base_score
            self.objective = objective_name_map.get(xgb_loader.name_obj, None)
            self.tree_output = tree_output_name_map.get(xgb_loader.name_obj, None)
            self.tree_limit = getattr(model, "best_ntree_limit", None)
            if xgb_loader.num_class > 0:
                self.num_stacked_models = xgb_loader.num_class
            # self.scaling = 1
        elif safe_isinstance(model, "xgboost.sklearn.XGBRanker"):
            import xgboost
            self.original_model = model.get_booster()
            self.model_type = "xgboost"
            xgb_loader = XGBTreeModelLoader(self.original_model)
            self.trees = xgb_loader.get_trees(data=data, data_missing=data_missing)
            self.base_offset = xgb_loader.base_score
            # Note: for ranker, leaving tree_output and objective as None as they
            # are not implemented in native code yet
            self.tree_limit = getattr(model, "best_ntree_limit", None)
            if xgb_loader.num_class > 0:
                self.num_stacked_models = xgb_loader.num_class
            # self.scaling = 1
        elif safe_isinstance(model, "lightgbm.basic.Booster"):
            assert_import("lightgbm")
            self.model_type = "lightgbm"
            self.original_model = model
            tree_info = self.original_model.dump_model()["tree_info"]
            try:
                self.trees = [SingleTree(e, data=data, data_missing=data_missing) for e in tree_info]
            except:
                self.trees = None  # we get here because the cext can't handle categorical splits yet

            self.objective = objective_name_map.get(model.params.get("objective", "regression"), None)
            self.tree_output = tree_output_name_map.get(model.params.get("objective", "regression"), None)
            # self.scaling = 1

        elif safe_isinstance(model, "gpboost.basic.Booster"):
            assert_import("gpboost")
            self.model_type = "gpboost"
            self.original_model = model
            tree_info = self.original_model.dump_model()["tree_info"]
            try:
                self.trees = [SingleTree(e, data=data, data_missing=data_missing) for e in tree_info]
            except:
                self.trees = None  # we get here because the cext can't handle categorical splits yet

            self.objective = objective_name_map.get(model.params.get("objective", "regression"), None)
            self.tree_output = tree_output_name_map.get(model.params.get("objective", "regression"), None)
            # self.scaling = 1

        elif safe_isinstance(model, "lightgbm.sklearn.LGBMRegressor"):
            assert_import("lightgbm")
            self.model_type = "lightgbm"
            self.original_model = model.booster_
            tree_info = self.original_model.dump_model()["tree_info"]
            try:
                self.trees = [SingleTree(e, data=data, data_missing=data_missing) for e in tree_info]
            except:
                self.trees = None  # we get here because the cext can't handle categorical splits yet
            self.objective = objective_name_map.get(model.objective, None)
            self.tree_output = tree_output_name_map.get(model.objective, None)
            if model.objective is None:
                self.objective = "squared_error"
                self.tree_output = "raw_value"
            # self.scaling = 1
        elif safe_isinstance(model, "lightgbm.sklearn.LGBMRanker"):
            assert_import("lightgbm")
            self.model_type = "lightgbm"
            self.original_model = model.booster_
            tree_info = self.original_model.dump_model()["tree_info"]
            try:
                self.trees = [SingleTree(e, data=data, data_missing=data_missing) for e in tree_info]
            except:
                self.trees = None  # we get here because the cext can't handle categorical splits yet
            # Note: for ranker, leaving tree_output and objective as None as they
            # are not implemented in native code yet
            # self.scaling = 1
        elif safe_isinstance(model, "lightgbm.sklearn.LGBMClassifier"):
            assert_import("lightgbm")
            self.model_type = "lightgbm"
            if model.n_classes_ > 2:
                self.num_stacked_models = model.n_classes_
            self.original_model = model.booster_
            tree_info = self.original_model.dump_model()["tree_info"]
            try:
                self.trees = [SingleTree(e, data=data, data_missing=data_missing) for e in tree_info]
            except:
                self.trees = None  # we get here because the cext can't handle categorical splits yet
            self.objective = objective_name_map.get(model.objective, None)
            self.tree_output = tree_output_name_map.get(model.objective, None)
            if model.objective is None:
                self.objective = "binary_crossentropy"
                self.tree_output = "log_odds"
            # self.scaling = 1
        elif safe_isinstance(model, "catboost.core.CatBoostRegressor"):
            assert_import("catboost")
            self.model_type = "catboost"
            self.original_model = model
            self.cat_feature_indices = model.get_cat_feature_indices()
            # self.scaling = 1
        elif safe_isinstance(model, "catboost.core.CatBoostClassifier"):
            assert_import("catboost")
            self.model_type = "catboost"
            self.original_model = model
            self.input_dtype = np.float32
            try:
                cb_loader = CatBoostTreeModelLoader(model)
                self.trees = cb_loader.get_trees(data=data, data_missing=data_missing)
            except:
                self.trees = None  # we get here because the cext can't handle categorical splits yet
            self.tree_output = "log_odds"
            self.objective = "binary_crossentropy"
            self.cat_feature_indices = model.get_cat_feature_indices()
            # self.scaling = 1
        elif safe_isinstance(model, "catboost.core.CatBoost"):
            assert_import("catboost")
            self.model_type = "catboost"
            self.original_model = model
            self.cat_feature_indices = model.get_cat_feature_indices()
            # self.scaling = 1
        elif safe_isinstance(model, "imblearn.ensemble._forest.BalancedRandomForestClassifier"):
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            self.trees = [SingleTree(e.tree_, normalize=True, scaling=scaling, data=data, data_missing=data_missing) for
                          e in model.estimators_]
            self.objective = objective_name_map.get(model.criterion, None)
            self.tree_output = "probability"
            # self.scaling = 1
        elif safe_isinstance(model, "ngboost.ngboost.NGBoost") or safe_isinstance(model,
                                                                                  "ngboost.api.NGBRegressor") or safe_isinstance(
            model, "ngboost.api.NGBClassifier"):
            assert model.base_models, "The NGBoost model has empty `base_models`! Have you called `model.fit`?"
            if self.model_output == "raw":
                param_idx = 0  # default to the first parameter of the output distribution
                warnings.warn(
                    "Translating model_ouput=\"raw\" to model_output=0 for the 0-th parameter in the distribution. Use model_output=0 directly to avoid this warning.")
            elif type(self.model_output) is int:
                param_idx = self.model_output
                self.model_output = "raw"  # note that after loading we have a new model_output type
            assert safe_isinstance(model.base_models[0][param_idx], ["sklearn.tree.DecisionTreeRegressor",
                                                                     "sklearn.tree.tree.DecisionTreeRegressor"]), "You must use default_tree_learner!"
            shap_trees = [trees[param_idx] for trees in model.base_models]
            self.internal_dtype = shap_trees[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = - model.learning_rate * np.array(model.scalings)  # output is weighted average of trees
            self.trees = [SingleTree(e.tree_, scaling=s, data=data, data_missing=data_missing) for e, s in
                          zip(shap_trees, scaling)]
            self.objective = objective_name_map.get(shap_trees[0].criterion, None)
            self.tree_output = "raw_value"
            self.base_offset = model.init_params[param_idx]
            # self.scaling = 1
        else:
            raise Exception("Model type not yet supported by TreeExplainer: " + str(type(model)))

        # build a dense numpy version of all the tree objects
        if self.trees is not None and self.trees:
            max_nodes = np.max([len(t.values) for t in self.trees])
            assert len(np.unique([t.values.shape[1] for t in
                                  self.trees])) == 1, "All trees in the ensemble must have the same output dimension!"
            num_trees = len(self.trees)
            if self.num_stacked_models > 1:
                assert len(
                    self.trees) % self.num_stacked_models == 0, "Only stacked models with equal numbers of trees are supported!"
                assert self.trees[0].values.shape[
                           1] == 1, "Only stacked models with single outputs per model are supported!"
                self.num_outputs = self.num_stacked_models
            else:
                self.num_outputs = self.trees[0].values.shape[1]

            if safe_isinstance(model, ["xgboost.sklearn.XGBClassifier",
                                       "catboost.core.CatBoostClassifier", "lightgbm.sklearn.LGBMClassifier",
                                       "sklearn.ensemble.GradientBoostingClassifier",
                                       "sklearn.ensemble._gb.GradientBoostingClassifier",
                                       "sklearn.ensemble.gradient_boosting.GradientBoostingClassifier"]) and \
                    self.num_outputs == 1:
                self.values_binary = np.zeros((num_trees, max_nodes, 2), dtype=self.internal_dtype)

                for i in range(num_trees):
                    # y = self.model.predict(self.data)
                    # self.trees[i].values = np.zeros((max_nodes, self.num_outputs))
                    # rebuild_acvtree(0, self.trees[i], self.data, y)
                    # self.trees[i].values = self.trees[i].scaling * self.trees[i].values
                    # p = np.exp(self.trees[i].values)/(1+np.exp(self.trees[i].values))
                    p = 1 / (1 + np.exp(-self.trees[i].values))
                    self.values_binary[i, :len(self.trees[i].values)] = np.concatenate([1 - p, p], axis=1) / num_trees

            # important to be -1 in unused sections!! This way we can tell which entries are valid.
            self.children_left = -np.ones((num_trees, max_nodes), dtype=np.int32)
            self.children_right = -np.ones((num_trees, max_nodes), dtype=np.int32)
            self.children_default = -np.ones((num_trees, max_nodes), dtype=np.int32)
            self.features = -np.ones((num_trees, max_nodes), dtype=np.int32)

            self.thresholds = np.zeros((num_trees, max_nodes), dtype=self.internal_dtype)
            self.values = np.zeros((num_trees, max_nodes, self.num_outputs), dtype=self.internal_dtype)
            self.node_sample_weight = np.zeros((num_trees, max_nodes), dtype=self.internal_dtype)

            self.partition_leaves_trees = []
            self.node_idx_trees = []
            self.data_leaves_trees = []
            self.leaf_idx_trees = []
            self.leaves_nb = []
            self.scalings = []
            for i in tqdm(range(num_trees)):
                self.scalings.append(self.trees[i].scaling)
                self.children_left[i, :len(self.trees[i].children_left)] = self.trees[i].children_left
                self.children_right[i, :len(self.trees[i].children_right)] = self.trees[i].children_right
                self.children_default[i, :len(self.trees[i].children_default)] = self.trees[i].children_default
                self.features[i, :len(self.trees[i].features)] = self.trees[i].features
                self.thresholds[i, :len(self.trees[i].thresholds)] = self.trees[i].thresholds
                if self.num_stacked_models > 1:
                    # stack_pos = int(i // (num_trees / self.num_stacked_models))
                    stack_pos = i % self.num_stacked_models
                    self.values[i, :len(self.trees[i].values[:, 0]), stack_pos] = self.trees[i].values[:, 0]
                else:
                    self.values[i, :len(self.trees[i].values)] = self.trees[i].values
                self.node_sample_weight[i, :len(self.trees[i].node_sample_weight)] = self.trees[i].node_sample_weight

                # ensure that the passed background dataset lands in every leaf
                if np.min(self.trees[i].node_sample_weight) <= 0:
                    self.fully_defined_weighting = False

                self.leaf_idx = [idx for idx in range(len(self.trees[i].features))
                                 if self.trees[i].children_left[idx] < 0]

                self.leaves_nb.append(len(self.leaf_idx))

                self.partition_leaves = []
                self.node_idx = []
                self.max_var = []
                self.data_leaves = []
                for leaf_id in self.leaf_idx:
                    node_id = [-1]
                    partition_leaf = [np.array([[-np.inf, np.inf]]) for idx2 in range(self.data.shape[1])]
                    _ = get_partition(leaf_id, partition_leaf, node_id, self.trees[i].children_left,
                                      self.trees[i].children_right, self.trees[i].features, self.trees[i].thresholds)

                    self.partition_leaves.append(np.squeeze(np.array(partition_leaf)))
                    self.node_idx.append(list(set(node_id[1:])))
                    self.max_var.append(len(self.node_idx[-1]))
                    # self.data_leaves.append(np.array([(self.data[:, s] <= self.partition_leaves[-1][s, 1]) * \
                    #                                       (self.data[:, s] > self.partition_leaves[-1][s, 0])
                    #                                       for s in range(self.data.shape[1])], dtype=np.int).transpose())

                self.partition_leaves_trees.append(self.partition_leaves)
                # self.data_leaves_trees.append(self.data_leaves)

                self.node_idx_trees.append(self.node_idx)
                self.leaf_idx_trees.append(self.leaf_idx)

            leaf_idx_trees = -np.ones(shape=(len(self.leaves_nb), np.max(self.leaves_nb)), dtype=np.int)
            partition_leaves_trees = -np.ones(
                shape=(len(self.leaves_nb), np.max(self.leaves_nb), self.data.shape[1], 2))
            # data_leaves_trees = -np.ones(shape=(len(self.leaves_nb), np.max(self.leaves_nb), self.data.shape[0], self.data.shape[1]), dtype=np.int)
            for i in range(len(self.leaves_nb)):
                leaf_idx_trees[i, :self.leaves_nb[i]] = np.array(self.leaf_idx_trees[i], dtype=np.int)
                partition_leaves_trees[i, :self.leaves_nb[i]] = np.array(self.partition_leaves_trees[i])
                # data_leaves_trees[i, :self.leaves_nb[i]] = np.array(self.data_leaves_trees[i], dtype=np.int)

            self.leaf_idx_trees = leaf_idx_trees
            self.partition_leaves_trees = partition_leaves_trees
            self.leaves_nb = np.array(self.leaves_nb, dtype=np.int)
            self.scalings = np.array(self.scalings, dtype=np.float)
            self.data = np.array(self.data, dtype=np.float)
            self.max_var = np.max(self.max_var)
            # self.data_leaves_trees = data_leaves_trees

            # if safe_isinstance(model, ["xgboost.sklearn.XGBClassifier",
            #                            "catboost.core.CatBoostClassifier", "lightgbm.sklearn.LGBMClassifier"]) and \
            #         self.num_outputs == 1:
            #     p = np.exp(self.values)/(1 + np.exp(self.values))
            #     print(np.max(p), np.min(1-p))
            #     self.values = np.concatenate([1-p, p], axis=2)
            #     self.num_outputs = 2

            self.num_nodes = np.array([len(t.values) for t in self.trees], dtype=np.int32)
            self.max_depth = np.max([t.max_depth for t in self.trees])

            if self.cache:
                if self.multi_threads:
                    self.lm, self.lm_s, self.lm_si = self.leaves_cache(C=self.C)
                else:
                    self.lm, self.lm_s, self.lm_si = self.leaves_cache_nopa(C=self.C)

                if self.cache_normalized:
                    if self.multi_threads:
                        self.lm_n, self.lm_s_n, self.lm_si_n = self.leaves_cache_normalized(C=self.C)
                    else:
                        self.lm_n, self.lm_s_n, self.lm_si_n = self.leaves_cache_normalized_nopa(C=self.C)

            # make sure the base offset is a 1D array
            if not hasattr(self.base_offset, "__len__") or len(self.base_offset) == 0:
                self.base_offset = (np.ones(self.num_outputs) * self.base_offset).astype(self.internal_dtype)
            self.base_offset = self.base_offset.flatten()
            assert len(self.base_offset) == self.num_outputs

    @abstractmethod
    def compute_cond_exp(self, X, S, data):
        pass

    @abstractmethod
    def shap_values(self, x, C):
        pass

    @abstractmethod
    def shap_values_acv(self, x, C, S_star, N_star):
        pass

    @abstractmethod
    def compute_sdp_clf(self, X, tX, S, data):
        pass

    @abstractmethod
    def compute_sdp_reg(self, X, tX, S, data):
        pass

    @abstractmethod
    def compute_local_sdp_clf(self, x, threshold, proba, index, data, final_coal, decay, C, verbose):
        pass

    @abstractmethod
    def compute_local_sdp_reg(self, x, threshold, proba, index, data, final_coal, decay, C, verbose):
        pass

    @abstractmethod
    def swing_values_clf(self, x, tx, S, data, threshold):
        pass

    @abstractmethod
    def swing_values_reg(self, x, tx, S, data, threshold):
        pass

    @abstractmethod
    def shap_values_swing_clf(self, x, tx, data, threshold, C):
        pass

    @abstractmethod
    def shap_values_swing_reg(self, x, tx, data, threshold, C):
        pass

    @abstractmethod
    def global_sdp_importance_clf(self, data, data_bground, columns_names, pi_level, decay, threshold,
                                  proba, C, verbose):
        pass

    @abstractmethod
    def global_sdp_importance_reg(self, data, data_bground, columns_names, pi_level, decay, threshold,
                                  proba, C, verbose):
        pass

    def predict(self, X, y=None, output=None, tree_limit=None):
        """ A consistent interface to make predictions from this model.

        Parameters
        ----------
        tree_limit : None (default) or int
            Limit the number of trees used by the model. By default None means no use the limit of the
            original model, and -1 means no limit.
        """

        if output is None:
            output = self.model_output

        if self.model_type == "pyspark":
            # import pyspark
            # TODO: support predict for pyspark
            raise NotImplementedError(
                "Predict with pyspark isn't implemented. Don't run 'interventional' as feature_perturbation.")

        # see if we have a default tree_limit in place.
        if tree_limit is None:
            tree_limit = -1 if self.tree_limit is None else self.tree_limit

        # convert dataframes
        if safe_isinstance(X, "pandas.core.series.Series"):
            X = X.values
        elif safe_isinstance(X, "pandas.core.frame.DataFrame"):
            X = X.values
        flat_output = False
        if len(X.shape) == 1:
            flat_output = True
            X = X.reshape(1, X.shape[0])
        if X.dtype.type != self.input_dtype:
            X = X.astype(self.input_dtype)
        X_missing = np.isnan(X, dtype=np.bool)
        assert isinstance(X, np.ndarray), "Unknown instance type: " + str(type(X))
        assert len(X.shape) == 2, "Passed input data matrix X must have 1 or 2 dimensions!"

        if tree_limit < 0 or tree_limit > self.values.shape[0]:
            tree_limit = self.values.shape[0]

        if output == "logloss":
            assert y is not None, "Both samples and labels must be provided when explaining the loss (i.e. `explainer.shap_values(X, y)`)!"
            assert X.shape[0] == len(
                y), "The number of labels (%d) does not match the number of samples to explain (%d)!" % (
                len(y), X.shape[0])
        #         transform = self.get_transform()
        assert_import("cext_acv")
        output = np.zeros((X.shape[0], self.num_outputs))
        cext_acv.dense_tree_predict(
            self.children_left, self.children_right, self.children_default,
            self.features, self.thresholds, self.values,
            self.max_depth, tree_limit, self.base_offset,
            X, X_missing, output)

        # drop dimensions we don't need
        if flat_output:
            if self.num_outputs == 1:
                return output.flatten()[0]
            else:
                return output.reshape(-1, self.num_outputs)
        else:
            if self.num_outputs == 1:
                return output.flatten()
            else:
                return output

    # def single_predict(self, X, y=None, output=None, tree_limit=None, i=0):
    #     """ A consistent interface to make predictions from this model.
    #
    #     Parameters
    #     ----------
    #     tree_limit : None (default) or int
    #         Limit the number of trees used by the model. By default None means no use the limit of the
    #         original model, and -1 means no limit.
    #     """
    #
    #     if output is None:
    #         output = self.model_output
    #
    #     if self.model_type == "pyspark":
    #         # import pyspark
    #         # TODO: support predict for pyspark
    #         raise NotImplementedError(
    #             "Predict with pyspark isn't implemented. Don't run 'interventional' as feature_perturbation.")
    #
    #     # see if we have a default tree_limit in place.
    #     if tree_limit is None:
    #         tree_limit = -1 if self.tree_limit is None else self.tree_limit
    #
    #     # convert dataframes
    #     if safe_isinstance(X, "pandas.core.series.Series"):
    #         X = X.values
    #     elif safe_isinstance(X, "pandas.core.frame.DataFrame"):
    #         X = X.values
    #     flat_output = False
    #     if len(X.shape) == 1:
    #         flat_output = True
    #         X = X.reshape(1, X.shape[0])
    #     if X.dtype.type != self.input_dtype:
    #         X = X.astype(self.input_dtype)
    #     X_missing = np.isnan(X, dtype=np.bool)
    #     assert isinstance(X, np.ndarray), "Unknown instance type: " + str(type(X))
    #     assert len(X.shape) == 2, "Passed input data matrix X must have 1 or 2 dimensions!"
    #
    #     if tree_limit < 0 or tree_limit > self.values.shape[0]:
    #         tree_limit = self.values.shape[0]
    #
    #     if output == "logloss":
    #         assert y is not None, "Both samples and labels must be provided when explaining the loss (i.e. `explainer.shap_values(X, y)`)!"
    #         assert X.shape[0] == len(
    #             y), "The number of labels (%d) does not match the number of samples to explain (%d)!" % (
    #             len(y), X.shape[0])
    #     #         transform = self.get_transform()
    #     assert_import("cext_acv")
    #     output = np.zeros((X.shape[0], self.num_outputs))
    #     cext_acv.single_tree_predict(
    #         self.children_left, self.children_right, self.children_default,
    #         self.features, self.thresholds, self.values,
    #         self.max_depth, tree_limit, self.base_offset,
    #         X, X_missing, output, i)
    #
    #     # drop dimensions we don't need
    #     if flat_output:
    #         if self.num_outputs == 1:
    #             return output.flatten()[0]/self.scaling
    #         else:
    #             return output.reshape(-1, self.num_outputs)/self.scaling
    #     else:
    #         if self.num_outputs == 1:
    #             return output.flatten()/self.scaling
    #         else:
    #             return output/self.scaling

    # def shap_values(self, x, C=[[]]):
    #     out = np.zeros((x.shape[0], x.shape[1], self.num_outputs))
    #     for i in range(len(self.trees)):
    #         out += shap_values_leaves(x, self.partition_leaves_trees[i], self.data_leaves_trees[i], self.node_idx_trees[i],
    #                                   self.leaf_idx_trees[i], self.node_sample_weight[i], self.values[i], C, self.num_outputs)
    #     return out
    #
    # def shap_values_acv(self, x, C=[[]]):
    #     out = np.zeros((x.shape[0], x.shape[1], self.num_outputs))
    #     for i in range(len(self.trees)):
    #         out += shap_values_leaves(x, self.partition_leaves_trees[i], self.data_leaves_trees[i], self.node_idx_trees[i],
    #                                   self.leaf_idx_trees[i], self.node_sample_weight[i], self.values[i], C, self.num_outputs)
    #     return out


class SingleTree:
    """ A single decision tree.

    The primary point of this object is to parse many different tree types into a common format.
    """

    def __init__(self, tree, normalize=False, scaling=1.0, data=None, data_missing=None):
        self.scaling = scaling
        if safe_isinstance(tree, ["sklearn.tree._tree.Tree", "econml.tree._tree.Tree", "skranger.tree._tree.Tree"]):
            self.children_left = tree.children_left.astype(np.int32)
            self.children_right = tree.children_right.astype(np.int32)
            self.children_default = self.children_left  # missing values not supported in sklearn
            self.features = tree.feature.astype(np.int32)
            self.thresholds = tree.threshold.astype(np.float64)
            self.values = tree.value.reshape(tree.value.shape[0], tree.value.shape[1] * tree.value.shape[2])
            if normalize:
                self.values = (self.values.T / self.values.sum(1)).T
            self.values = self.values * scaling
            self.node_sample_weight = tree.weighted_n_node_samples.astype(np.float64)

        elif type(tree) is dict and 'features' in tree:
            self.children_left = tree["children_left"].astype(np.int32)
            self.children_right = tree["children_right"].astype(np.int32)
            self.children_default = tree["children_default"].astype(np.int32)
            self.features = tree["features"].astype(np.int32)
            self.thresholds = tree["thresholds"]
            self.values = tree["values"] * scaling
            self.node_sample_weight = tree["node_sample_weight"]

        # deprecated dictionary support (with sklearn singlular style "feature" and "value" names)
        elif type(tree) is dict and 'children_left' in tree:
            self.children_left = tree["children_left"].astype(np.int32)
            self.children_right = tree["children_right"].astype(np.int32)
            self.children_default = tree["children_default"].astype(np.int32)
            self.features = tree["feature"].astype(np.int32)
            self.thresholds = tree["threshold"]
            self.values = tree["value"] * scaling
            self.node_sample_weight = tree["node_sample_weight"]

        elif safe_isinstance(tree, "pyspark.ml.classification.DecisionTreeClassificationModel") \
                or safe_isinstance(tree, "pyspark.ml.regression.DecisionTreeRegressionModel"):
            # model._java_obj.numNodes() doesn't give leaves, need to recompute the size
            def getNumNodes(node, size):
                size = size + 1
                if node.subtreeDepth() == 0:
                    return size
                else:
                    size = getNumNodes(node.leftChild(), size)
                    return getNumNodes(node.rightChild(), size)

            num_nodes = getNumNodes(tree._java_obj.rootNode(), 0)
            self.children_left = np.full(num_nodes, -2, dtype=np.int32)
            self.children_right = np.full(num_nodes, -2, dtype=np.int32)
            self.children_default = np.full(num_nodes, -2, dtype=np.int32)
            self.features = np.full(num_nodes, -2, dtype=np.int32)
            self.thresholds = np.full(num_nodes, -2, dtype=np.float64)
            self.values = [-2] * num_nodes
            self.node_sample_weight = np.full(num_nodes, -2, dtype=np.float64)

            def buildTree(index, node):
                index = index + 1
                if tree._java_obj.getImpurity() == 'variance':
                    self.values[index] = [node.prediction()]  # prediction for the node
                else:
                    self.values[index] = [e for e in
                                          node.impurityStats().stats()]  # for gini: NDarray(numLabel): 1 per label: number of item for each label which went through this node
                self.node_sample_weight[
                    index] = node.impurityStats().count()  # weighted count of element trough this node

                if node.subtreeDepth() == 0:
                    return index
                else:
                    self.features[
                        index] = node.split().featureIndex()  # index of the feature we split on, not available for leaf, int
                    if str(node.split().getClass()).endswith('tree.CategoricalSplit'):
                        # Categorical split isn't implemented, TODO: could fake it by creating a fake node to split on the exact value?
                        raise NotImplementedError('CategoricalSplit are not yet implemented')
                    self.thresholds[
                        index] = node.split().threshold()  # threshold for the feature, not available for leaf, float

                    self.children_left[index] = index + 1
                    idx = buildTree(index, node.leftChild())
                    self.children_right[index] = idx + 1
                    idx = buildTree(idx, node.rightChild())
                    return idx

            buildTree(-1, tree._java_obj.rootNode())
            # default Not supported with mlib? (TODO)
            self.children_default = self.children_left
            self.values = np.asarray(self.values)
            if normalize:
                self.values = (self.values.T / self.values.sum(1)).T
            self.values = self.values * scaling

        elif type(tree) == dict and 'tree_structure' in tree:  # LightGBM model dump
            start = tree['tree_structure']
            num_parents = tree['num_leaves'] - 1
            self.children_left = np.empty((2 * num_parents + 1), dtype=np.int32)
            self.children_right = np.empty((2 * num_parents + 1), dtype=np.int32)
            self.children_default = np.empty((2 * num_parents + 1), dtype=np.int32)
            self.features = np.empty((2 * num_parents + 1), dtype=np.int32)
            self.thresholds = np.empty((2 * num_parents + 1), dtype=np.float64)
            self.values = [-2] * (2 * num_parents + 1)
            self.node_sample_weight = np.empty((2 * num_parents + 1), dtype=np.float64)
            visited, queue = [], [start]
            while queue:
                vertex = queue.pop(0)
                if 'split_index' in vertex.keys():
                    if vertex['split_index'] not in visited:
                        if 'split_index' in vertex['left_child'].keys():
                            self.children_left[vertex['split_index']] = vertex['left_child']['split_index']
                        else:
                            self.children_left[vertex['split_index']] = vertex['left_child']['leaf_index'] + num_parents
                        if 'split_index' in vertex['right_child'].keys():
                            self.children_right[vertex['split_index']] = vertex['right_child']['split_index']
                        else:
                            self.children_right[vertex['split_index']] = vertex['right_child'][
                                                                             'leaf_index'] + num_parents
                        if vertex['default_left']:
                            self.children_default[vertex['split_index']] = self.children_left[vertex['split_index']]
                        else:
                            self.children_default[vertex['split_index']] = self.children_right[vertex['split_index']]
                        self.features[vertex['split_index']] = vertex['split_feature']
                        self.thresholds[vertex['split_index']] = vertex['threshold']
                        self.values[vertex['split_index']] = [vertex['internal_value']]
                        self.node_sample_weight[vertex['split_index']] = vertex['internal_count']
                        visited.append(vertex['split_index'])
                        queue.append(vertex['left_child'])
                        queue.append(vertex['right_child'])
                else:
                    self.children_left[vertex['leaf_index'] + num_parents] = -1
                    self.children_right[vertex['leaf_index'] + num_parents] = -1
                    self.children_default[vertex['leaf_index'] + num_parents] = -1
                    self.features[vertex['leaf_index'] + num_parents] = -1
                    self.children_left[vertex['leaf_index'] + num_parents] = -1
                    self.children_right[vertex['leaf_index'] + num_parents] = -1
                    self.children_default[vertex['leaf_index'] + num_parents] = -1
                    self.features[vertex['leaf_index'] + num_parents] = -1
                    self.thresholds[vertex['leaf_index'] + num_parents] = -1
                    self.values[vertex['leaf_index'] + num_parents] = [vertex['leaf_value']]
                    self.node_sample_weight[vertex['leaf_index'] + num_parents] = vertex['leaf_count']
            self.values = np.asarray(self.values)
            self.values = np.multiply(self.values, scaling)

        elif type(tree) == dict and 'nodeid' in tree:
            """ Directly create tree given the JSON dump (with stats) of a XGBoost model.
            """

            def max_id(node):
                if "children" in node:
                    return max(node["nodeid"], *[max_id(n) for n in node["children"]])
                else:
                    return node["nodeid"]

            m = max_id(tree) + 1
            self.children_left = -np.ones(m, dtype=np.int32)
            self.children_right = -np.ones(m, dtype=np.int32)
            self.children_default = -np.ones(m, dtype=np.int32)
            self.features = -np.ones(m, dtype=np.int32)
            self.thresholds = np.zeros(m, dtype=np.float64)
            self.values = np.zeros((m, 1), dtype=np.float64)
            self.node_sample_weight = np.empty(m, dtype=np.float64)

            def extract_data(node, tree):
                i = node["nodeid"]
                tree.node_sample_weight[i] = node["cover"]

                if "children" in node:
                    tree.children_left[i] = node["yes"]
                    tree.children_right[i] = node["no"]
                    tree.children_default[i] = node["missing"]
                    tree.features[i] = node["split"]
                    tree.thresholds[i] = node["split_condition"]

                    for n in node["children"]:
                        extract_data(n, tree)
                elif "leaf" in node:
                    tree.values[i] = node["leaf"] * scaling

            extract_data(tree, self)

        elif type(tree) == str:
            """ Build a tree from a text dump (with stats) of xgboost.
            """

            nodes = [t.lstrip() for t in tree[:-1].split("\n")]
            nodes_dict = {}
            for n in nodes: nodes_dict[int(n.split(":")[0])] = n.split(":")[1]
            m = max(nodes_dict.keys()) + 1
            children_left = -1 * np.ones(m, dtype="int32")
            children_right = -1 * np.ones(m, dtype="int32")
            children_default = -1 * np.ones(m, dtype="int32")
            features = -2 * np.ones(m, dtype="int32")
            thresholds = -1 * np.ones(m, dtype="float64")
            values = 1 * np.ones(m, dtype="float64")
            node_sample_weight = np.zeros(m, dtype="float64")
            values_lst = list(nodes_dict.values())
            keys_lst = list(nodes_dict.keys())
            for i in range(0, len(keys_lst)):
                value = values_lst[i]
                key = keys_lst[i]
                if ("leaf" in value):
                    # Extract values
                    val = float(value.split("leaf=")[1].split(",")[0])
                    node_sample_weight_val = float(value.split("cover=")[1])
                    # Append to lists
                    values[key] = val
                    node_sample_weight[key] = node_sample_weight_val
                else:
                    c_left = int(value.split("yes=")[1].split(",")[0])
                    c_right = int(value.split("no=")[1].split(",")[0])
                    c_default = int(value.split("missing=")[1].split(",")[0])
                    feat_thres = value.split(" ")[0]
                    if ("<" in feat_thres):
                        feature = int(feat_thres.split("<")[0][2:])
                        threshold = float(feat_thres.split("<")[1][:-1])
                    if ("=" in feat_thres):
                        feature = int(feat_thres.split("=")[0][2:])
                        threshold = float(feat_thres.split("=")[1][:-1])
                    node_sample_weight_val = float(value.split("cover=")[1].split(",")[0])
                    children_left[key] = c_left
                    children_right[key] = c_right
                    children_default[key] = c_default
                    features[key] = feature
                    thresholds[key] = threshold
                    node_sample_weight[key] = node_sample_weight_val

            self.children_left = children_left
            self.children_right = children_right
            self.children_default = children_default
            self.features = features
            self.thresholds = thresholds
            self.values = values[:, np.newaxis] * scaling
            self.node_sample_weight = node_sample_weight
        else:
            raise Exception("Unknown input to SingleTree constructor: " + str(tree))

        # Re-compute the number of samples that pass through each node if we are given data
        # if data is not None and data_missing is not None:
        #     self.node_sample_weight[:] = 0.0
        #     cext_acv.dense_tree_update_weights(
        #         self.children_left, self.children_right, self.children_default, self.features,
        #         self.thresholds, self.values, 1, self.node_sample_weight, data, data_missing
        #     )

        # we compute the expectations to make sure they follow the SHAP logic
        self.max_depth = cext_acv.compute_expectations(
            self.children_left, self.children_right, self.node_sample_weight,
            self.values
        )

    def predict(self, X):
        # see if we have a default tree_limit in place.
        children_left = np.expand_dims(self.children_left, 0)
        children_right = np.expand_dims(self.children_right, 0)
        children_default = np.expand_dims(self.children_default, 0)
        features = np.expand_dims(self.features, 0)
        thresholds = np.expand_dims(self.thresholds, 0)
        values = np.expand_dims(self.values, 0) / self.scaling
        # node_sample_weight = np.expand_dims(self.node_sample_weight, 0)
        # convert dataframes
        if safe_isinstance(X, "pandas.core.series.Series"):
            X = X.values
        elif safe_isinstance(X, "pandas.core.frame.DataFrame"):
            X = X.values
        flat_output = False
        if len(X.shape) == 1:
            flat_output = True
            X = X.reshape(1, X.shape[0])
        # if X.dtype.type != self.input_dtype:
        #     X = X.astype(self.input_dtype)
        X_missing = np.isnan(X, dtype=np.bool)
        assert isinstance(X, np.ndarray), "Unknown instance type: " + str(type(X))
        assert len(X.shape) == 2, "Passed input data matrix X must have 1 or 2 dimensions!"

        assert_import("cext_acv")

        tree_limit = 1
        self.num_outputs = self.values.shape[1]
        base_offset = np.zeros(self.num_outputs)
        output = np.zeros((X.shape[0], self.num_outputs))
        cext_acv.dense_tree_predict(
            children_left, children_right, children_default, features, thresholds, values,
            self.max_depth, tree_limit, base_offset, X, X_missing, output)

        # drop dimensions we don't need
        if flat_output:
            if self.num_outputs == 1:
                return output.flatten()[0]
            else:
                return output.reshape(-1, self.num_outputs)
        else:
            if self.num_outputs == 1:
                return output.flatten()
            else:
                return output


class IsoTree(SingleTree):
    """
    In sklearn the tree of the Isolation Forest does not calculated in a good way.
    """

    def __init__(self, tree, tree_features, normalize=False, scaling=1.0, data=None, data_missing=None):
        super(IsoTree, self).__init__(tree, normalize, scaling, data, data_missing)
        if safe_isinstance(tree, "sklearn.tree._tree.Tree"):
            from sklearn.ensemble._iforest import _average_path_length  # pylint: disable=no-name-in-module

            def _recalculate_value(tree, i, level):
                if tree.children_left[i] == -1 and tree.children_right[i] == -1:
                    value = level + _average_path_length(np.array([tree.n_node_samples[i]]))[0]
                    self.values[i, 0] = value
                    return value * tree.n_node_samples[i]
                else:
                    value_left = _recalculate_value(tree, tree.children_left[i], level + 1)
                    value_right = _recalculate_value(tree, tree.children_right[i], level + 1)
                    self.values[i, 0] = (value_left + value_right) / tree.n_node_samples[i]
                    return value_left + value_right

            _recalculate_value(tree, 0, 0)
            if normalize:
                self.values = (self.values.T / self.values.sum(1)).T
            self.values = self.values * scaling
            # re-number the features if each tree gets a different set of features
            self.features = np.where(self.features >= 0, tree_features[self.features], self.features)


def get_xgboost_json(model):
    """ This gets a JSON dump of an XGBoost model while ensuring the features names are their indexes.
    """
    fnames = model.feature_names
    model.feature_names = None
    json_trees = model.get_dump(with_stats=True, dump_format="json")
    model.feature_names = fnames

    # this fixes a bug where XGBoost can return invalid JSON
    json_trees = [t.replace(": inf,", ": 1000000000000.0,") for t in json_trees]
    json_trees = [t.replace(": -inf,", ": -1000000000000.0,") for t in json_trees]

    return json_trees


class XGBTreeModelLoader(object):
    """ This loads an XGBoost model directly from a raw memory dump.

    We can't use the JSON dump because due to numerical precision issues those
    tree can actually be wrong when feature values land almost on a threshold.
    """

    def __init__(self, xgb_model):
        # new in XGBoost 1.1, 'binf' is appended to the buffer
        self.buf = xgb_model.save_raw().lstrip(b'binf')
        self.pos = 0

        # load the model parameters
        self.base_score = self.read('f')
        self.num_feature = self.read('I')
        self.num_class = self.read('i')
        self.contain_extra_attrs = self.read('i')
        self.contain_eval_metrics = self.read('i')
        self.read_arr('i', 29)  # reserved
        self.name_obj_len = self.read('Q')
        self.name_obj = self.read_str(self.name_obj_len)
        self.name_gbm_len = self.read('Q')
        self.name_gbm = self.read_str(self.name_gbm_len)

        # new in XGBoost 1.0 is that the base_score is saved untransformed (https://github.com/dmlc/xgboost/pull/5101)
        # so we have to transform it depending on the objective
        import xgboost
        if LooseVersion(xgboost.__version__).version[0] >= 1:
            if self.name_obj in ["binary:logistic", "reg:logistic"]:
                self.base_score = scipy.special.logit(self.base_score)  # pylint: disable=no-member

        assert self.name_gbm == "gbtree", "Only the 'gbtree' model type is supported, not '%s'!" % self.name_gbm

        # load the gbtree specific parameters
        self.num_trees = self.read('i')
        self.num_roots = self.read('i')
        self.num_feature = self.read('i')
        self.pad_32bit = self.read('i')
        self.num_pbuffer_deprecated = self.read('Q')
        self.num_output_group = self.read('i')
        self.size_leaf_vector = self.read('i')
        self.read_arr('i', 32)  # reserved

        # load each tree
        self.num_roots = np.zeros(self.num_trees, dtype=np.int32)
        self.num_nodes = np.zeros(self.num_trees, dtype=np.int32)
        self.num_deleted = np.zeros(self.num_trees, dtype=np.int32)
        self.max_depth = np.zeros(self.num_trees, dtype=np.int32)
        self.num_feature = np.zeros(self.num_trees, dtype=np.int32)
        self.size_leaf_vector = np.zeros(self.num_trees, dtype=np.int32)
        self.node_parents = []
        self.node_cleft = []
        self.node_cright = []
        self.node_sindex = []
        self.node_info = []
        self.loss_chg = []
        self.sum_hess = []
        self.base_weight = []
        self.leaf_child_cnt = []
        for i in range(self.num_trees):

            # load the per-tree params
            self.num_roots[i] = self.read('i')
            self.num_nodes[i] = self.read('i')
            self.num_deleted[i] = self.read('i')
            self.max_depth[i] = self.read('i')
            self.num_feature[i] = self.read('i')
            self.size_leaf_vector[i] = self.read('i')

            # load the nodes
            self.read_arr('i', 31)  # reserved
            self.node_parents.append(np.zeros(self.num_nodes[i], dtype=np.int32))
            self.node_cleft.append(np.zeros(self.num_nodes[i], dtype=np.int32))
            self.node_cright.append(np.zeros(self.num_nodes[i], dtype=np.int32))
            self.node_sindex.append(np.zeros(self.num_nodes[i], dtype=np.uint32))
            self.node_info.append(np.zeros(self.num_nodes[i], dtype=np.float32))
            for j in range(self.num_nodes[i]):
                self.node_parents[-1][j] = self.read('i')
                self.node_cleft[-1][j] = self.read('i')
                self.node_cright[-1][j] = self.read('i')
                self.node_sindex[-1][j] = self.read('I')
                self.node_info[-1][j] = self.read('f')

            # load the stat nodes
            self.loss_chg.append(np.zeros(self.num_nodes[i], dtype=np.float32))
            self.sum_hess.append(np.zeros(self.num_nodes[i], dtype=np.float32))
            self.base_weight.append(np.zeros(self.num_nodes[i], dtype=np.float32))
            self.leaf_child_cnt.append(np.zeros(self.num_nodes[i], dtype=np.int))
            for j in range(self.num_nodes[i]):
                self.loss_chg[-1][j] = self.read('f')
                self.sum_hess[-1][j] = self.read('f')
                self.base_weight[-1][j] = self.read('f')
                self.leaf_child_cnt[-1][j] = self.read('i')

    def get_trees(self, data=None, data_missing=None):
        shape = (self.num_trees, self.num_nodes.max())
        self.children_default = np.zeros(shape, dtype=np.int)
        self.features = np.zeros(shape, dtype=np.int)
        self.thresholds = np.zeros(shape, dtype=np.float32)
        self.values = np.zeros((shape[0], shape[1], 1), dtype=np.float32)
        trees = []
        for i in range(self.num_trees):
            for j in range(self.num_nodes[i]):
                if np.right_shift(self.node_sindex[i][j], np.uint32(31)) != 0:
                    self.children_default[i, j] = self.node_cleft[i][j]
                else:
                    self.children_default[i, j] = self.node_cright[i][j]
                self.features[i, j] = self.node_sindex[i][j] & ((np.uint32(1) << np.uint32(31)) - np.uint32(1))
                if self.node_cleft[i][j] >= 0:
                    # Xgboost uses < for thresholds where shap uses <=
                    # Move the threshold down by the smallest possible increment
                    self.thresholds[i, j] = np.nextafter(self.node_info[i][j], - np.float32(np.inf))
                else:
                    self.values[i, j] = self.node_info[i][j]

            l = len(self.node_cleft[i])
            trees.append(SingleTree({
                "children_left": self.node_cleft[i],
                "children_right": self.node_cright[i],
                "children_default": self.children_default[i, :l],
                "feature": self.features[i, :l],
                "threshold": self.thresholds[i, :l],
                "value": self.values[i, :l],
                "node_sample_weight": self.sum_hess[i]
            }, data=data, data_missing=data_missing))
        return trees

    def read(self, dtype):
        size = struct.calcsize(dtype)
        val = struct.unpack(dtype, self.buf[self.pos:self.pos + size])[0]
        self.pos += size
        return val

    def read_arr(self, dtype, n_items):
        format = "%d%s" % (n_items, dtype)
        size = struct.calcsize(format)
        val = struct.unpack(format, self.buf[self.pos:self.pos + size])[0]
        self.pos += size
        return val

    def read_str(self, size):
        val = self.buf[self.pos:self.pos + size].decode('utf-8')
        self.pos += size
        return val

    def print_info(self):

        print("--- global parmeters ---")
        print("base_score =", self.base_score)
        print("num_feature =", self.num_feature)
        print("num_class =", self.num_class)
        print("contain_extra_attrs =", self.contain_extra_attrs)
        print("contain_eval_metrics =", self.contain_eval_metrics)
        print("name_obj_len =", self.name_obj_len)
        print("name_obj =", self.name_obj)
        print("name_gbm_len =", self.name_gbm_len)
        print("name_gbm =", self.name_gbm)
        print()
        print("--- gbtree specific parameters ---")
        print("num_trees =", self.num_trees)
        print("num_roots =", self.num_roots)
        print("num_feature =", self.num_feature)
        print("pad_32bit =", self.pad_32bit)
        print("num_pbuffer_deprecated =", self.num_pbuffer_deprecated)
        print("num_output_group =", self.num_output_group)
        print("size_leaf_vector =", self.size_leaf_vector)


class CatBoostTreeModelLoader:
    def __init__(self, cb_model):
        # cb_model.save_model("cb_model.json", format="json")
        # self.loaded_cb_model = json.load(open("cb_model.json", "r"))
        import tempfile
        tmp_file = tempfile.NamedTemporaryFile()
        cb_model.save_model(tmp_file.name, format="json")
        self.loaded_cb_model = json.load(open(tmp_file.name, "r"))
        tmp_file.close()

        # load the CatBoost oblivious trees specific parameters
        self.num_trees = len(self.loaded_cb_model['oblivious_trees'])
        self.max_depth = self.loaded_cb_model['model_info']['params']['tree_learner_options']['depth']

    def get_trees(self, data=None, data_missing=None):
        # load each tree
        trees = []
        for tree_index in range(self.num_trees):

            # load the per-tree params
            # depth = len(self.loaded_cb_model['oblivious_trees'][tree_index]['splits'])

            # load the nodes

            # Re-compute the number of samples that pass through each node if we are given data
            leaf_weights = self.loaded_cb_model['oblivious_trees'][tree_index]['leaf_weights']
            leaf_weights_unraveled = [0] * (len(leaf_weights) - 1) + leaf_weights
            leaf_weights_unraveled[0] = sum(leaf_weights)
            for index in range(len(leaf_weights) - 2, 0, -1):
                leaf_weights_unraveled[index] = leaf_weights_unraveled[2 * index + 1] + leaf_weights_unraveled[
                    2 * index + 2]

            leaf_values = self.loaded_cb_model['oblivious_trees'][tree_index]['leaf_values']
            leaf_values_unraveled = [0] * (len(leaf_values) - 1) + leaf_values

            children_left = [i * 2 + 1 for i in range(len(leaf_values) - 1)]
            children_left += [-1] * len(leaf_values)

            children_right = [i * 2 for i in range(1, len(leaf_values))]
            children_right += [-1] * len(leaf_values)

            children_default = [i * 2 + 1 for i in range(len(leaf_values) - 1)]
            children_default += [-1] * len(leaf_values)

            # load the split features and borders
            # split features and borders go from leafs to the root
            split_features_index = []
            borders = []

            # split features and borders go from leafs to the root
            for elem in self.loaded_cb_model['oblivious_trees'][tree_index]['splits']:
                split_type = elem.get('split_type')
                if split_type == 'FloatFeature':
                    split_feature_index = elem.get('float_feature_index')
                    borders.append(elem['border'])
                elif split_type == 'OneHotFeature':
                    split_feature_index = elem.get('cat_feature_index')
                    borders.append(elem['value'])
                else:
                    split_feature_index = elem.get('ctr_target_border_idx')
                    borders.append(elem['border'])
                split_features_index.append(split_feature_index)

            split_features_index_unraveled = []
            for counter, feature_index in enumerate(split_features_index[::-1]):
                split_features_index_unraveled += [feature_index] * (2 ** counter)
            split_features_index_unraveled += [0] * len(leaf_values)

            borders_unraveled = []
            for counter, border in enumerate(borders[::-1]):
                borders_unraveled += [border] * (2 ** counter)
            borders_unraveled += [0] * len(leaf_values)

            trees.append(SingleTree({"children_left": np.array(children_left),
                                     "children_right": np.array(children_right),
                                     "children_default": np.array(children_default),
                                     "feature": np.array(split_features_index_unraveled),
                                     "threshold": np.array(borders_unraveled),
                                     "value": np.array(leaf_values_unraveled).reshape((-1, 1)),
                                     "node_sample_weight": np.array(leaf_weights_unraveled),
                                     }, data=data, data_missing=data_missing))

        return trees



from .base_agnostree import *
from .py_acv import *
from .utils_sdp import *
from .utils import *
import numpy as np
import cyext_acv, cyext_acv_nopa, cyext_acv_cache
from acv_explainers.utils import extend_partition
from skranger.ensemble import RangerForestClassifier, RangerForestRegressor
from sklearn.utils.validation import check_is_fitted, check_X_y, column_or_1d, check_array, as_float_array, \
    check_consistent_length
from sklearn.exceptions import NotFittedError
import random

class ACXplainer:
    def __init__(
            self,
            classifier,
            n_estimators=50,
            verbose=False,
            mtry=0,
            importance="impurity",
            min_node_size=0,
            max_depth=5,
            replace=True,
            sample_fraction=None,
            keep_inbag=False,
            inbag=None,
            split_rule="gini",
            num_random_splits=1,
            seed=2021,
            quantiles=None
    ):
        """
        ACXplainer is a agnostic explainer that computes two different explanations for any model or data:

                -  Same Decision Probability and Sufficient Explanations
                -  Minimal Sufficient Rules
        It can also fit a Global Explainable Rules which is an interpretable model by design.

        Args:
            classifier (bool): If True, the ACXplainer is in classification mode, otherwise regression

            n_estimators (int): Number of trees used in the forest

            verbose (bool): Enable ranger's verbose logging

            mtry (int/callable): The number of features to split on each node. When a
                        callable is passed, the function must accept a single parameter which is the
                        number of features passed, and return some value between 1 and the number of
                        features.

            importance (string):  One of one of ``none``, ``impurity``, ``impurity_corrected``, ``permutation``.

            min_node_size (int): The minimal node size.

            max_depth (int): The maximal tree depth; 0 means unlimited.

            replace (bool): Sample with replacement.

            sample_fraction (float): The fraction of observations to sample. The default is 1 when sampling with
                                     replacement, and 0.632 otherwise. This can be a list of class specific values.
            keep_inbag (bool): If true, save how often observations are in-bag in each tree.

            inbag (list): A list of size ``n_estimators``, containing inbag counts for each
                          observation. Can be used for stratified sampling.

            split_rule (string): One of 'gini', 'extratrees', 'hellinger'; default 'gini' for classification and
                                one of 'variance', 'extratrees', 'maxstat', 'beta' for regression; default 'variance'

            num_random_splits (int): The number of random splits to consider for the ``extratrees`` splitrule.

            seed (int): Random seed value
        """
        self.classifier = int(classifier)
        self.n_estimators = n_estimators
        self.verbose = verbose
        self.mtry = mtry
        self.importance = importance
        self.min_node_size = min_node_size
        self.max_depth = max_depth
        self.replace = replace
        self.sample_fraction = sample_fraction
        self.keep_inbag = keep_inbag
        self.inbag = inbag
        self.split_rule = split_rule
        self.num_random_splits = num_random_splits
        self.check_is_explain = False
        self.ACXplainer = None
        self.seed = seed
        self.rules = None
        self.rules_output = None
        self.rules_s_star = None
        self.rules_coverage = None
        self.rules_acc = None
        self.rules_var = None
        self.d = None
        self.check_is_globalrule = False
        self.rules_output_proba = None
        self.rules_ori = None
        self.rules_s_star_ori = None

        if self.classifier:
            self.model = RangerForestClassifier(self.n_estimators,
                                                self.verbose,
                                                self.mtry,
                                                self.importance,
                                                self.min_node_size,
                                                self.max_depth,
                                                self.replace,
                                                self.sample_fraction,
                                                self.keep_inbag,
                                                self.inbag,
                                                self.split_rule,
                                                self.num_random_splits,
                                                seed=self.seed,
                                                enable_tree_details=True)
        else:
            if self.split_rule == 'gini':
                self.split_rule = 'variance'
            self.model = RangerForestRegressor(self.n_estimators,
                                               self.verbose,
                                               self.mtry,
                                               self.importance,
                                               self.min_node_size,
                                               self.max_depth,
                                               self.replace,
                                               self.sample_fraction,
                                               self.keep_inbag,
                                               self.inbag,
                                               self.split_rule,
                                               self.num_random_splits,
                                               seed=self.seed,
                                               enable_tree_details=True,
                                               quantiles=True)

    def fit(self, X, y, sample_weight=None, split_select_weights=None, always_split_features=None,
            categorical_features=None):
        """
        Fit the Random forest estimator that will be used for the different estimation using the trainind data for the explanations

        Args:
            X (numpy.ndarray): training features (# samples X # features)

            y (numpy.ndarray): training targets (# samples)

            sample_weight (numpy.ndarray): Optional weights for input samples

            split_select_weights (list): Optional Vector of weights between 0 and 1 of probabilities to select features
                                         for splitting. Can be a single vector or a vector of vectors with one vector per tree.

            always_split_features (list): Features which should always be selected for splitting. A list of column
                                          index values.

            categorical_features (list): A list of column index values which should be considered categorical, or unordered.
        """
        X, y = check_X_y(X, y, dtype=np.double)
        self.d = X.shape[1]
        self.model.fit(X=X, y=y, sample_weight=sample_weight, split_select_weights=split_select_weights,
                       always_split_features=always_split_features, categorical_features=categorical_features)

    def fit_global_rules(self, X, y, rules, rules_s_star):
        # TODO: think about removing s_star ? add complexity ?
        X, y = check_X_y(X, y, dtype=np.double)
        y = as_float_array(y)
        self.rules_ori = rules
        self.rules_s_star_ori = rules_s_star
        self.rules, self.rules_s_star = unique_rules_s_star(rules, rules_s_star)
        if self.classifier:
            self.rules_coverage, self.rules_acc, self.rules_var, self.rules_output, self.rules_output_proba = \
                compute_rules_metrics(
                    self.rules,
                    X,
                    y,
                    self.rules_s_star,
                    self.classifier)
        else:
            self.rules_coverage, self.rules_acc, self.rules_var, self.rules_output = compute_rules_metrics(self.rules,
                                                                                                           X,
                                                                                                           y,
                                                                                                           self.rules_s_star,
                                                                                                           self.classifier)
        self.check_is_globalrule = True

    def predict(self, X):
        X = check_array(X)
        check_is_fitted(self.model,
                        msg="This ACXplainer instance is not fitted yet. Call 'fit' with appropriate arguments"
                            " before using this estimator")
        return self.model.predict(X)

    def predict_proba(self, X):
        X = check_array(X)
        check_is_fitted(self.model,
                        msg="This ACXplainer instance is not fitted yet. Call 'fit' with appropriate arguments"
                            " before using this estimator")
        if self.classifier:
            return self.model.predict_proba(X)

    def predict_global_rules(self, X, coverage=0.0, min_acc=0.8, min_reg=1000):
        if not self.check_is_globalrule:
            raise NotFittedError("The global Explainable Rules is not fitted yet. Call 'fit_global_rules' "
                                 "with appropriate arguments")
        X = check_array(X, dtype=np.double)
        if self.classifier:
            return global_rules_model(X, self.rules, self.rules_output, self.rules_coverage,
                                      self.rules_acc, coverage, self.rules_s_star, min_acc)
        else:
            return global_rules_model_reg(X, self.rules, self.rules_output, self.rules_coverage,
                                          self.rules_acc, coverage, self.rules_s_star, min_reg)

    def predict_proba_global_rules(self, X, coverage=0.0, min_acc=0.8):
        if not self.check_is_globalrule:
            raise NotFittedError("The global Explainable Rules is not fitted yet. Call 'fit_global_rules' "
                                 "with appropriate arguments")
        if not self.classifier:
            raise ValueError('Predict_proba function is only available for classification')

        X = check_array(X, dtype=np.double)
        return global_rules_model(X, self.rules, self.rules_output_proba, self.rules_coverage,
                                  self.rules_acc, coverage, self.rules_s_star, min_acc)

    def check_is_explainer(self):
        check_is_fitted(self.model,
                        msg="This ACXplainer instance is not fitted yet. Call 'fit' with appropriate arguments"
                            " before using this estimator")
        if not self.check_is_explain:
            self.ACXplainer = BaseAgnosTree(self.model, self.d)
            self.check_is_explain = True

    def compute_sdp_rf(self, X, y, data, y_data, S, min_node_size=5, t=20):
        """
         Estimate the Same Decision Probability (SDP) of a set of samples X given subset S using the consistent estimator
         (Projected Forest + Quantile Regression), see Paper [ref]

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the SDP

            y (numpy.ndarray): 1-D array (# samples) the targets of X

            data (numpy.ndarray): The background dataset that is used for the estimation of the SDP. It should be the
                                training samples.

            y_data (numpy.ndarray): The targets of the background dataset

            S (list[list[int]]): A list that contains the column indices of the variable on which we condition to compute the SDP
                                 for each observation

            min_node_size (int): The minimal node size of the Projected Random Forest

            t (float):  The level of variations around the prediction that defines the SDP in regression (only for regression)

        Returns:
            sdp (numpy.ndarray):  A 1-D matrix (# samples), sdp[i] contains the Same Decision Probability (SDP)
                                 of the Minimal Sufficient Explanation of observation i
        """
        X, y = check_X_y(X, y, dtype=np.double)
        data, y_data = check_X_y(data, y_data, dtype=np.double)
        y, y_data = as_float_array(y).astype(np.double), as_float_array(y_data).astype(np.double)
        try:
            check_consistent_length(X, S)
        except ValueError as exp:
            raise ValueError('{} for X (samples) and S (coalition)'.format(exp))

        self.check_is_explainer()

        sdp = cyext_acv.compute_sdp_rf(X, y, data, y_data, S, self.ACXplainer.features, self.ACXplainer.thresholds,
                                       self.ACXplainer.children_left,
                                       self.ACXplainer.children_right, self.ACXplainer.max_depth, min_node_size,
                                       self.classifier, t)
        return sdp

    def compute_sdp_rule(self, X, y, data, y_data, S, min_node_size=5, t=20):
        """
         Estimate the local rule-based explanations of a set of samples X given Sufficient Explanations S using
         the consistent estimator (Projected Forest + Quantile Regression), see Paper [ref]. For each observation x, all
         observations that falls in the rule of x has the same SDP as x.

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the SDP

            y (numpy.ndarray): 1-D array (# samples) the targets of X

            data (numpy.ndarray): The background dataset that is used for the estimation of the SDP. It should be the
                                training samples.

            y_data (numpy.ndarray): The targets of the background dataset

            S (list[list[int]]): A list that contains the indices of the columns of the Sufficient Explanation for each observation

            min_node_size (int): The minimal node size of the Projected Random forest

            t (float): The level of variations around the prediction that defines the SDP in regression (only for regression)

        Returns:
            sdp (numpy.ndarray): A 1-D matrix (# samples), sdp[i] contains the Same Decision Probability (SDP)
                                 of the Sufficient Explanation of observation i
            rules (numpy.ndarray): A matrix (# samples x # features x 2), rules[i] that contains the local rule of observation i
        """
        X, y = check_X_y(X, y, dtype=[np.double, np.double])
        data, y_data = check_X_y(data, y_data, dtype=[np.double, np.double])
        y, y_data = as_float_array(y).astype(np.double), as_float_array(y_data).astype(np.double)
        try:
            check_consistent_length(X, S)
        except ValueError as exp:
            raise ValueError('{} for X (samples) and S (coalition)'.format(exp))

        self.check_is_explainer()

        sdp, rules = cyext_acv.compute_sdp_rule(X, y, data, y_data, S, self.ACXplainer.features,
                                                self.ACXplainer.thresholds,
                                                self.ACXplainer.children_left,
                                                self.ACXplainer.children_right, self.ACXplainer.max_depth,
                                                min_node_size, self.classifier, t)
        return sdp, rules

    def compute_sdp_maxrules(self, X, y, data, y_data, S, min_node_size=5, t=20, pi_level=0.95,
                             verbose=False, algo2=False, rdm_size=20):
        """
         Estimate the maximal rule-based explanations of a set of samples X given Sufficient Explanations S using the consistent estimator
         (Projected Forest + Quantile Regression), see Paper [ref]. For each instance x, all
         observations that falls in the rule of x has its SDP given S above \pi as x.

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the SDP

            y (numpy.ndarray): 1-D array (# samples) the targets of X

            data (numpy.ndarray): The background dataset that is used for the estimation of the SDP. It should be the
                                training samples.

            y_data (numpy.ndarray): The targets of the background dataset

            S (list[list[int]]): A list that contains the indices of the columns of the variable of the Sufficient
                                Explanations for each observation

            min_node_size (int): The minimal node size of the Projected Forest

            t (double): The level of variations around the prediction that defines the SDP in regression (only for regression)

            pi_level (double): The minimal level of SDP for the Sufficient Explanations

        Returns:
            sdp (numpy.ndarray): A 1-D matrix (# samples), sdp[i] contains the Same Decision Probability (SDP)
                                 of the Sufficient Explanation of observation i
            rules (numpy.ndarray): A matrix (# samples x # features x 2), rules[i] that contains the maximal local rule
                                   of observation i
        """
        X, y = check_X_y(X, y, dtype=np.double)
        data, y_data = check_X_y(data, y_data, dtype=np.double)
        y, y_data = as_float_array(y).astype(np.double), as_float_array(y_data).astype(np.double)
        try:
            check_consistent_length(X, S)
        except ValueError as exp:
            raise ValueError('{} for X (samples) and S (coalition)'.format(exp))

        self.check_is_explainer()

        if not verbose:
            sdp, rules, sdp_all, rules_data, w = cyext_acv.compute_sdp_maxrule(X, y, data, y_data, S,
                                                                               self.ACXplainer.features,
                                                                               self.ACXplainer.thresholds,
                                                                               self.ACXplainer.children_left,
                                                                               self.ACXplainer.children_right,
                                                                               self.ACXplainer.max_depth,
                                                                               min_node_size, self.classifier, t,
                                                                               pi_level)
        else:
            sdp, rules, sdp_all, rules_data, w = cyext_acv.compute_sdp_maxrule_verbose(X, y, data, y_data, S,
                                                                                       self.ACXplainer.features,
                                                                                       self.ACXplainer.thresholds,
                                                                                       self.ACXplainer.children_left,
                                                                                       self.ACXplainer.children_right,
                                                                                       self.ACXplainer.max_depth,
                                                                                       min_node_size, self.classifier,
                                                                                       t, pi_level)

        if not algo2:
            extend_partition(rules, rules_data, sdp_all, pi=pi_level, S=S)
        else:
            # rules_rdm = [rules.copy() for z in range(rdm_size)]
            # for i in tqdm(range(rdm_size)):
            #     for a, ru in enumerate(rules_rdm[i]):
            #         rules_comp = [rules_data[a, j] for j in range(rules_data.shape[1]) if sdp_all[a, j] >= sdp[a]]
            #         random.shuffle(rules_comp)
            #         r, _ = return_best(rules_rdm[i][a], rules_comp, S[a])
            #         rules_rdm[i][a][S[a]] = r
            for a in tqdm(range(rules.shape[0])):
                rules_sets = [rules_data[a, j] for j in range(rules_data.shape[1]) if sdp_all[a, j] >= sdp[a]]
                rules[a][S[a]] = overapprox_rule(rules[a], rules_sets, S[a])
            # return sdp, rules_rdm, sdp_all, rules_data, w

        return sdp, rules, sdp_all, rules_data, w

    def importance_sdp_rf(self, X, y, data, y_data, min_node_size=5, t=20,
                          C=[[]], pi_level=0.9, minimal=1, stop=True):
        """
         Estimate the Minimal Sufficient Explanations and the Global sdp importance of a set of samples X using the
         consistent estimator (Projected Forest + Quantile Regression), see Paper [ref].

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the SDP

            y (numpy.ndarray): 1-D array (# samples) the targets of X

            data (numpy.ndarray): The background dataset that is used for the estimation of the SDP. It should be the
                                training samples.

            y_data (numpy.ndarray): The targets of the background dataset

            min_node_size (int): The minimal node size

            t (float): The level of variations around the prediction that defines the SDP in regression (only for regression)

            C (list[list[int]]): A list that contains a list of the indices of column for each grouped variables

            pi_level (float): The minimal value of the Same Decision Probability (SDP) of the Sufficient Explanations.
                              It should be in (0, 1).

            minimal (int): It will search the Sufficient Explanations from subsets of size "minimal" instead of 1 by default

            stop (bool): If stop=True, it will stop searching for the Sufficient Explanations if it does not find
                         any Sufficient Explanations smaller than (# features / 2), otherwise it will continues until
                         end.

        Returns:
            global_sdp_importance (numpy.ndarray): A 1-D matrix (# features) that is the global explanatory importance
                                                 based on samples X. For a given i, sdp_importance[i] corresponds to the
                                                 frequency of apparition of feature i in the Minimal Sufficient Explanations
                                                 of the set of samples X

            sdp_index (numpy.ndarray): A matrix (# samples X # features) that contains the indices of the variables in the
                                     the Minimal Sufficient Explanations for each sample. For a given i, the positive
                                     values of sdp_index[i] corresponds to the Minimal Sufficient Explanations of
                                     observation i.

            size (numpy.ndarray): A 1-D matrix (# samples) that contains the size of the Minimal Sufficient Explanation
                                for each sample.

            sdp (numpy.ndarray): A 1-D matrix (# samples) that contains the Same Decision Probability (SDP)
                               of the Sufficient Explanation for each sample.
        """
        X, y = check_X_y(X, y, dtype=np.double)
        data, y_data = check_X_y(data, y_data, dtype=np.double)
        y, y_data = as_float_array(y).astype(np.double), as_float_array(y_data).astype(np.double)
        self.check_is_explainer()

        if X.shape[1] > 10:
            feature_importance = -np.array(self.model.feature_importances_)
            search_space = list(np.argsort(feature_importance))
            # flat_list = [item for t in self.ACXplainer.node_idx_trees for sublist in t for item in sublist]
            # node_idx = pd.Series(flat_list)
            # search_space = []
            # for v in (node_idx.value_counts().keys()):
            #     search_space += [v]
        else:
            search_space = [i for i in range(X.shape[1])]

        return cyext_acv.global_sdp_rf(X, y, data, y_data, self.ACXplainer.features, self.ACXplainer.thresholds,
                                       self.ACXplainer.children_left,
                                       self.ACXplainer.children_right, self.ACXplainer.max_depth, min_node_size,
                                       self.classifier, t, C,
                                       pi_level, minimal, stop, search_space[:10])

    def sufficient_expl_rf(self, X, y, data, y_data, min_node_size=5, t=20,
                           C=[[]], pi_level=0.9, minimal=1, stop=True):

        """
         Estimate all the Sufficient Explanations and the Global sdp importance of a set of samples X using the
         consistent estimator (Projected Forest + Quantile Regression), see Paper [ref].

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the SDP

            y (numpy.ndarray): 1-D array (# samples) the targets of X

            data (numpy.ndarray): The background dataset that is used for the estimation of the SDP. It should be the
                                training samples.

            y_data (numpy.ndarray): The targets of the background dataset

            min_node_size (int): The minimal node size

            t (float): The level of variations around the prediction that defined the SDP for regression (only for regression)

            C (list[list[int]]): A list that contains a list of the indices of the column for each grouped variables

            pi_level (float): The minimal value of the Same Decision Probability (SDP) of the Sufficient Explanations.
                              It should be in (0, 1).

            minimal (int): It will search the Sufficient Explanations from subsets of size "minimal" instead of 1 by default

            stop (bool): If stop=True, it will stop searching for the Sufficient Explanations if it does not find
                         any Sufficient Explanations smaller than (# features / 2), otherwise it will continues until
                         end.

        Returns:
            sufficient_coal (list[list[list[int]]]): a list that contains the column indices of the Sufficient Explanations
                                                     of each sample
            sdp_coal (list[list[list[int]]]): a list that contains the SDP of the Sufficient Explanations of each sample

            sdp_global (numpy.ndarray): 1-D array (# features), sdp_global[i] corresponds to the frequency of apparition
            of variable i in the **all** the sufficient explanations over the samples X.
        """

        X, y = check_X_y(X, y, dtype=np.double)
        data, y_data = check_X_y(data, y_data, dtype=np.double)
        y, y_data = as_float_array(y).astype(np.double), as_float_array(y_data).astype(np.double)
        self.check_is_explainer()
        if X.shape[1] > 10:

            feature_importance = -np.array(self.model.feature_importances_)
            search_space = list(np.argsort(feature_importance))
            # flat_list = [item for t in self.ACXplainer.node_idx_trees for sublist in t for item in sublist]
            # node_idx = pd.Series(flat_list)
            # search_space = []
            # for v in (node_idx.value_counts().keys()):
            #     search_space += [v]
        else:
            search_space = [i for i in range(X.shape[1])]
        # TODO: Remove the [-1] in the Sufficent Coal
        return cyext_acv.sufficient_expl_rf(X, y, data, y_data, self.ACXplainer.features, self.ACXplainer.thresholds,
                                            self.ACXplainer.children_left,
                                            self.ACXplainer.children_right, self.ACXplainer.max_depth, min_node_size,
                                            self.classifier, t, C,
                                            pi_level, minimal, stop, search_space[:10])

    def compute_exp_rf(self, X, y, data, y_data, S, min_node_size=5, t=20):
        """
         Estimate the Conditional Expectation (exp) of a set of samples X given subset S using the consistent estimator
         (Projected Forest), see Paper [ref]

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the SDP

            y (numpy.ndarray): 1-D array (# samples) the targets of X

            data (numpy.ndarray): The background dataset that is used for the estimation of the SDP. It should be the
                                training samples.

            y_data (numpy.ndarray): The targets of the background dataset

            S (list[list[int]]): A list that contains the indices of the variable on which we condition to compute the SDP
                                 for each observation

            min_node_size (int): The minimal node size

            t (float): not used here

        Returns:
            exp (numpy.ndarray): A 1-D matrix (# samples), exp[i] contains the Conditional Expectation of observation i
                                 given S
        """
        X, y = check_X_y(X, y, dtype=np.double)
        data, y_data = check_X_y(data, y_data, dtype=np.double)
        y, y_data = as_float_array(y).astype(np.double), as_float_array(y_data).astype(np.double)
        try:
            check_consistent_length(X, S)
        except ValueError as exp:
            raise ValueError('{} for X (samples) and S (coalition)'.format(exp))

        self.check_is_explainer()

        exp = cyext_acv.compute_exp_rf(X, y, data, y_data, S, self.ACXplainer.features, self.ACXplainer.thresholds,
                                       self.ACXplainer.children_left,
                                       self.ACXplainer.children_right, self.ACXplainer.max_depth, min_node_size,
                                       self.classifier, t)
        return exp

    def compute_cdf_rf(self, X, y, data, y_data, S, min_node_size=5, t=20):
        """
         Estimate the Projected Cumulative Distribution Function (P-CDF) of a set of samples X given subset S using the
          consistent estimator (Projected Forest + Quantile Regression), see Paper [ref]

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the SDP

            y (numpy.ndarray): 1-D array (# samples) the targets of X

            data (numpy.ndarray): The background dataset that is used for the estimation of the SDP. It should be the
                                training samples.

            y_data (numpy.ndarray): The targets of the background dataset

            S (list[list[int]]): A list that contains the indices of the variable on which we condition to compute the P-CDF
                                 for each observation

            min_node_size (int): The minimal node size of the Projected Forest

            t (float): not used here

        Returns:
            exp (numpy.ndarray): A 1-D matrix (# samples), exp[i] contains the Conditional Expectation of observation i
                                 given S
        """
        X, y = check_X_y(X, y, dtype=np.double)
        data, y_data = check_X_y(data, y_data, dtype=np.double)
        y, y_data = as_float_array(y).astype(np.double), as_float_array(y_data).astype(np.double)
        try:
            check_consistent_length(X, S)
        except ValueError as exp:
            raise ValueError('{} for X (samples) and S (coalition)'.format(exp))

        self.check_is_explainer()

        sdp = cyext_acv.compute_cdf_rf(X, y, data, y_data, S, self.ACXplainer.features, self.ACXplainer.thresholds,
                                       self.ACXplainer.children_left,
                                       self.ACXplainer.children_right, self.ACXplainer.max_depth, min_node_size,
                                       self.classifier, t)
        return sdp

    @staticmethod
    def compute_local_sdp(d, sufficient_coal):
        """
        Estimate the local explanatory importance of each variable given the set of Sufficient Explanations of each
        sample
        Args:
            d (int): # features
            sufficient_coal (list[list[list[int]]]): a list that contains the column indices of the Sufficient Explanations
                                                     of each sample

        Returns:
            local_exp_imp (numpy.ndarray): 1-D array (# features) the local explanatory importance of each variable for each
                                           sample.
        """
        flat = [item for sublist in sufficient_coal for item in sublist]
        flat = pd.Series(flat)
        flat = dict(flat.value_counts() / len(sufficient_coal))
        local_sdp = np.zeros(d)
        for key in flat.keys():
            local_sdp[key] = flat[key]
        return local_sdp

    @staticmethod
    def compute_msdp_clf(X, S, data, model=None):
        """
        Compute marginal SDP of a set of sample X given S for classifier
        """
        return msdp(X, S, model, data)

    @staticmethod
    def importance_msdp_clf_search(X, data, model=None, C=[[]], minimal=1, pi_level=0.9, r_search_space=None,
                                   stop=True):
        """
        Compute the marginal minimal sufficient explanations of any classifier
        """

        return importance_msdp_clf_search(X=X, rg_data=data, model=model, C=C, minimal=minimal,
                                          pi_level=pi_level, r_search_space=r_search_space, stop=stop)

    @staticmethod
    def compute_msdp_reg(X, S, data, model=None, threshold=0.2):
        """
         Compute marginal SDP of a set of sample X given S for regressor
        """
        return msdp_reg(X, S, model, data, threshold)

    @staticmethod
    def importance_msdp_reg_search(X, data, model=None, C=[[]], minimal=1, pi_level=0.9, threshold=0.2,
                                   r_search_space=None, stop=True):
        """
        Compute the marginal minimal sufficient explanations of any regressor model
        """
        return importance_msdp_reg_search(X, rg_data=data, model=model, C=C, minimal=minimal, pi_level=pi_level,
                                          threshold=threshold, r_search_space=r_search_space, stop=stop)

    # def compute_quantile_rf(self, X, y, data, y_data, S, min_node_size=5, classifier=1, t=20, quantile=95):
    #     X, y = check_X_y(X, y, dtype=np.double)
    #     data, y_data = check_X_y(data, y_data, dtype=np.double)
    #     y, y_data = as_float_array(y).astype(np.double), as_float_array(y_data).astype(np.double)
    #     try:
    #         check_consistent_length(X, S)
    #     except ValueError as exp:
    #         raise ValueError('{} for X (samples) and S (coalition)'.format(exp))
    #
    #     self.check_is_explainer()
    #
    #     # TODO: Optimize the loops
    #     y_quantiles = cyext_acv.compute_quantile_rf(X, y, data, y_data, S, self.ACXplainer.features,
    #                                                 self.ACXplainer.thresholds,
    #                                                 self.ACXplainer.children_left,
    #                                                 self.ACXplainer.children_right, self.ACXplainer.max_depth,
    #                                                 min_node_size, classifier, t,
    #                                                 quantile)
    #     return y_quantiles
    #
    # def compute_quantile_diff_rf(self, X, y, data, y_data, S, min_node_size=5, classifier=1, t=20, quantile=95):
    #     X, y = check_X_y(X, y, dtype=np.double)
    #     y, y_data = as_float_array(y).astype(np.double), as_float_array(y_data).astype(np.double)
    #     data, y_data = check_X_y(data, y_data, dtype=np.double)
    #     try:
    #         check_consistent_length(X, S)
    #     except ValueError as exp:
    #         raise ValueError('{} for X (samples) and S (coalition)'.format(exp))
    #
    #     # TODO: Optimize the loops
    #     self.check_is_explainer()
    #     y_quantiles_diff = cyext_acv.compute_quantile_diff_rf(X, y, data, y_data, S, self.ACXplainer.features,
    #                                                           self.ACXplainer.thresholds,
    #                                                           self.ACXplainer.children_left,
    #                                                           self.ACXplainer.children_right, self.ACXplainer.max_depth,
    #                                                           min_node_size,
    #                                                           classifier, t, quantile)
    #     return y_quantiles_diff



from acv_app.plots import *
import numpy as np
import streamlit as st
# from acv_explainers import ACVTree

def write_pg(x_train, x_test, y_train, y_test, acvtree):
    @st.cache(allow_output_mutation=True)
    def compute_sv():
        # treemodel = ACVTree(acvtree.model, x_train.values.astype(np.double))
        # return treemodel.shap_values(x_test.values.astype(np.double)[:100], C=[[]])
        acvtree.check_is_explainer()
        return acvtree.ACXplainer.shap_values(x_test.values.astype(np.double)[:100], x_train.values.astype(np.double))
    idx = st.selectbox(
        'Choose the observation you want to explain',
        list(range(x_test.shape[0]))
    )

    col1, col2 = st.columns(2)
    with col1:
        sv = compute_sv()
        if acvtree.classifier != True:
            pred = acvtree.predict(x_test.values.astype(np.double))
        else:
            pred = acvtree.predict_proba(x_test.values.astype(np.double))
        if sv.shape[2] != 1:
            base = np.mean(pred[:, 1])
            fig = waterfall(base, sv[idx, :, 1], x_test.values[idx], x_test.columns)
        else:
            base = np.mean(pred)
            fig = waterfall(base, sv[idx, :, 0], x_test.values[idx], x_test.columns)
        st.pyplot(fig)


from .utils import *

class ExperimentsLinearGMMOhe:

    def __init__(self, mean={'a': [0., 0., 0.], 'b': [0., 0., 0.], 'c': [0., 0., 0.]},
                 cov={key: 0.8 * np.ones(shape=(3, 3)) - (0.8 - 1) * np.eye(3) for key in ['a', 'b', 'c']},
                 coefs={'a': [1, 1, 1], 'b': [1, 1, 1], 'c': [1, 1, 1]},
                 pi={'a': 0.25, 'b': 0.25, 'c': 0.5},
                 S=[0, 1, 2, 'a', 'b'], n=1000, C=[[]], cat_index=[]):
        """
        Load toy experiments where data D={(X, Y, Z), i={1,...n}} are built with linear model f, such that Z = f(X, Y) = B_Y^t X,
        X follow a multivariate gaussian distribution N(mu, Sigma) and Y are in {a, b, c}.
        Categorical variable are encode with one hot encoding (OHE).

        Args:
            mean (dict[array]): a dict that contains the means for each class Y in {a, b, c}
            cov (dict[array]): a dict that contains the covariance matrix for each class Y in {a, b, c}
            coefs (dict[array]): a dict that contains coefficients of the linear regression for each class Y in {a, b, c}
            pi (dict[float]): a dict that contains the probabilities P(Y=y) for each class Y in {a, b, c}
            n (int): the size of training and test set
            C (list[list]): list of variables group together
            cat_index (list): index of the categorical variables
        """

        # Python init
        self.n = n
        self.d = len(mean['a'])
        self.index = list(np.arange(self.d))
        self.cat_index = cat_index
        self.index_cat = {cat_index[0]: 'a', cat_index[1]: 'b', cat_index[2]:'c'}
        self.coefs = coefs

        self.cov = cov
        self.mean = {key: np.array(mean[key]) for key in mean.keys()}
        self.C = C
        self.pi = pi

        self.dummy = {'a': np.array([1, 0, 0]), 'b': np.array([0, 1, 0]), 'c': np.array([0, 0, 0])}
        self.revers_dummy = {str([1., 0., 0.]): 'a', str([0., 1., 0.]): 'b', str([0., 0., 0.]): 'c'}

        self.gaus = {key: st.multivariate_normal(self.mean[key], self.cov[key]) for key in self.mean.keys()}

        # R init
        # self.r_mean = {key: FloatVector(self.mean[key]) for key in self.mean.keys()}
        # sig = {key: robjects.Vector(self.cov[key]) for key in self.cov.keys()}
        # self.r_cov = {key: robjects.Matrix(sig[key]) for key in sig.keys()}

        # Data init

        self.data, self.y_train = self.gen_data(n)
        self.data_test, self.y_test = self.gen_data(n)

    def gen_data_y(self, y='a', n=10000):
        """
        Generate data D={(X, Y, Z), i={1,...n}} of class Y

        Args:
            y (string): label of the class
            n (int): size of the sample

        Returns:
            array: X
            array: Z
        """
        data_gen = st.multivariate_normal(self.mean[y], self.cov[y])
        data = data_gen.rvs(n)
        y_train = linear_regression(self.coefs[y], data)
        return data, y_train

    def gen_data(self, n):
        """
        Generate data D={([X, Y], Z), i={1,...n}}

        Args:
            n (int): size of data

        Returns:
            array: X
            array: Z
        """

        mixture_idx = np.random.choice(list(self.pi.keys()), size=n, replace=True, p=list(self.pi.values()))
        data_a = self.gaus['a'].rvs(np.sum(mixture_idx == 'a'))
        y_a = linear_regression(self.coefs['a'], data_a)
        data_a = np.concatenate([data_a, np.tile([1, 0, 0], (np.sum(mixture_idx == 'a'), 1))], axis=1)

        data_b = self.gaus['b'].rvs(np.sum(mixture_idx == 'b'))
        y_b = linear_regression(self.coefs['b'], data_b)
        data_b = np.concatenate([data_b, np.tile([0, 1, 0], (np.sum(mixture_idx == 'b'), 1))], axis=1)

        data_c = self.gaus['c'].rvs(np.sum(mixture_idx == 'c'))
        y_c = linear_regression(self.coefs['c'], data_c)
        data_c = np.concatenate([data_c, np.tile([0, 0, 0], (np.sum(mixture_idx == 'c'), 1))], axis=1)

        data, y_train = np.concatenate([data_a, data_b, data_c], axis=0), np.concatenate([y_a, y_b, y_c], axis=0)
        return data, y_train

    def linear_gmm(self, x):
        """
        Compute f([X,Y]) for an observation x = [X, Y] where f is the linear model

        Args:
            x (array): [X, Y] observation where variable Y are encode with dummy variables

        Returns:
            float: f([X,Y])
        """
        a_cond = (x[self.cat_index[0]] == 1) * (x[self.cat_index[1]] == 0)
        b_cond = (x[self.cat_index[0]] == 0) * (x[self.cat_index[1]] == 1)
        c_cond = (x[self.cat_index[0]] == 0) * (x[self.cat_index[1]] == 0)

        return linear_regression_0(self.coefs['a'], x[:self.d]) * a_cond \
               + linear_regression_0(self.coefs['b'], x[:self.d]) * b_cond \
               + linear_regression_0(self.coefs['c'], x[:self.d]) * c_cond

    def linear_gmm_batch(self, x):
        """
        Compute f([X,Y]) for a batch of observation x = [X, Y] where f is the linear model

        Args:
            x (array): batch àf observations where variable Y are encode with dummy variables

        Returns:
            float: f([X,Y])
        """
        a_cond = (x[:, self.cat_index[0]] == 1) * (x[:, self.cat_index[1]] == 0)
        b_cond = (x[:, self.cat_index[0]] == 0) * (x[:, self.cat_index[1]] == 1)
        c_cond = (x[:, self.cat_index[0]] == 0) * (x[:, self.cat_index[1]] == 0)

        return linear_regression(self.coefs['a'], x[:, :self.d]) * \
               a_cond + linear_regression(self.coefs['b'], x[:, :self.d]) * b_cond \
               + linear_regression(self.coefs['c'], x[:, :self.d]) * c_cond

    def density_x_y_set(self, x, y, S):
        """
        Compute marginal density given S evaluate on x

        Args:
            x (array): observation
            y (string): label of the class Y in {a, b, c}
            S (list): index of the continuous variable on which we condition

        Returns:
            float: marginal density given S evaluate on x
        """

        # S doesn't take str (y)
        x_in = x.copy()[:self.d]
        if S != []:  # if condition on numeric variable S
            x_in = x_in[S]
            mean = self.mean[y][S]
            cov = self.cov[y][S][:, S]
        else:  # this loops is useless ?
            mean = self.mean[y]
            cov = self.cov[y]

        gaus_S = st.multivariate_normal(mean, cov)
        cond_dist_x = gaus_S.pdf(x_in)

        return cond_dist_x

    def mar_exp_linear_y_set(self, x, y, S):
        """"
        Compute exact marginal expectation given S and class y of the linear model E[f(x, X_{\bar{S}})]

        Args:
            x (array): observation
            y (string): label of the class {a, b, c}
            S (list): index of the continuous variable on which on condition

        Returns:
            float: marginal expectation given S and class y of the linear model E[f(x, X_{\bar{S}})]
        """

        # S doesn't take str (y)
        x_in = x.copy()[:self.d]

        S_bar = [i for i in self.index if i not in S]

        cond_dist_param = marMVN(self.mean[y], self.cov[y], S_bar, S, x_in[S])

        mu_bar = cond_dist_param[0]
        if len(S) != self.d:
            x_in[S_bar] = mu_bar

        out = np.sum(self.coefs[y] * x_in)
        return out

    def cond_exp_linear_y_set(self, x, y, S):
        """
        Compute exact conditional expectation given S and class y of the linear model E[f(x)| X_S=x_S, Y=y]

        Args:
            x (array): observation
            y (string): label of the class {a, b, c}
            S (list): index of the continuous variable on which on condition

        Returns:
            float: exact conditional expectation given S and class y of the linear model E[f(x)| X_S=x_S, Y=y]
        """

        # S doesn't take str (y)
        x_in = x.copy()[:self.d]

        S_bar = [i for i in self.index if i not in S]

        cond_dist_param = condMVN(self.mean[y], self.cov[y], S_bar, S, x_in[S])

        mu_bar = cond_dist_param[0]
        if len(S) != self.d:
            x_in[S_bar] = mu_bar

        out = np.sum(self.coefs[y] * x_in)
        return out

    def cond_exp_linear_gmm(self, x, S):
        """
        Compute exact conditional expectation given S of the linear model E[f(x)| X_S=x_S]. Here, S can contains the
        index of the categorical variable.

        Args:
            x (array): observation
            S (list): index of the variable (continuous and categorical) on which on condition

        Returns:
            float: conditional expectation given S and class y of the linear model E[f(x)| X_S=x_S]
        """

        x_in = x.copy()
        if len(S) != len(x_in):  # if not condition on all variable

            cat_va = [ca for ca in S if ca in self.cat_index]  # get categorical index
            S = [cu for cu in S if cu not in self.cat_index]  # get numerical index

            cond_exp = 0
            if cat_va == []:  # if not condition on categorical variable
                p_x = 0
                p_x_y = 0
                if S != []:  # if condition on numeric variable
                    for key in self.mean.keys():
                        p_x += self.pi[key] * self.density_x_y_set(x_in, key, S)
                        p_x_y += self.pi[key] * self.density_x_y_set(x_in, key, S) * self.cond_exp_linear_y_set(x_in,
                                                                                                                key, S)
                    cond_exp = p_x_y / p_x
                else:
                    for key in self.mean.keys():
                        cond_exp += self.pi[key] * self.cond_exp_linear_y_set(x_in, key, S)
            else:  # if condition on categorical variable
                if len(cat_va) >= 2:  # ==> condition on a group of dummy variable [y1, y2] ==> on cat variable
                    cond_exp = self.cond_exp_linear_y_set(x, self.revers_dummy[str(list(x[self.cat_index]))], S)

                else:  # ==> conditioning on 1 dummy
                    if 1. in x_in[cat_va]:
                        cond_exp = self.cond_exp_linear_y_set(x, self.revers_dummy[str(list(x[self.cat_index]))], S)
                    else:
                        cat_sample = [key for key in self.pi.keys() if key != self.index_cat[cat_va[0]]]
                        p_total = 0
                        if S != []:
                            for key in cat_sample:
                                p_total += self.pi[key] * self.density_x_y_set(x_in, key, S)
                                cond_exp += self.pi[key] * self.density_x_y_set(x_in, key,
                                                                                S) * self.cond_exp_linear_y_set(x,
                                                                                                                key,
                                                                                                                S)
                            cond_exp = cond_exp / p_total
                        else:
                            for key in cat_sample:
                                p_total += self.pi[key]
                                cond_exp += self.pi[key] * self.cond_exp_linear_y_set(x_in, key, S)
                            cond_exp = cond_exp / p_total
            return cond_exp
        else:
            return self.linear_gmm(x_in)

    def cond_exp_monte_carlo_linear(self, x, S, N):
        """
        Compute conditional expectation given S of the linear model with monte carlo estimator E[f(x)| X_S=x_S]. Here, S can contains the
        index of the categorical variable.

        Args:
            x (array): observation
            S (list): index of the variable (continuous and categorical) on which on condition
            N (int): size of the sampling of monte carlo estimator

        Returns:
            float: Conditional expectation given S of the linear model with monte carlo estimator E[f(x)| X_S=x_S]
        """

        x_in = x.copy()
        cat_va = [ca for ca in S if ca in self.cat_index]  # get categorical index
        S_cu = [cu for cu in S if cu not in self.cat_index]  # get numerical index

        def get_given_data(idx):
            val = np.array([x[idx]])
            val = np.tile(val, N)
            return val

        def gen_data_by_cat(cat_va, prob, N):
            # to do: make it real function of S, c_index, etc

            mixture_idx = np.random.choice(cat_va, size=N, replace=True, p=prob)

            rg = {key: sampleMVN(np.sum(mixture_idx == key), self.mean[key], self.cov[key], S_bar, S, x_in[S])
                  for key in cat_va if np.sum(mixture_idx == key) != 0}
            rg = np.concatenate([np.concatenate([rg[key], np.tile(self.dummy[key], (np.sum(mixture_idx == key), 1))],
                                                axis=1) for key in rg.keys()])

            rg_data = pd.DataFrame(rg, columns=[str(s) for s in S_bar] + [str(ca) for ca in self.cat_index])

            for val_id in S:
                rg_data[str(val_id)] = get_given_data(val_id)

            rg_data = rg_data[sorted(rg_data.columns)]
            return rg_data

        if len(S_cu) != self.d:
            cat_va = [ca for ca in S if ca in self.cat_index]  # get categorical index
            S = [cu for cu in S if cu not in self.cat_index]  # get numerical index
            S_bar = [i for i in self.index if i not in S]

            if cat_va == []:
                if S != []:
                    prob = [self.pi[key] * self.density_x_y_set(x_in, key, S) for key in self.pi.keys()]
                    p_x = np.sum(prob)
                    prob = [p / p_x for p in prob]
                    mixture_idx = np.random.choice(list(self.pi.keys()), size=N, replace=True, p=prob)
                else:
                    mixture_idx = np.random.choice(list(self.pi.keys()), size=N, replace=True, p=list(self.pi.values()))

                rg = {key: sampleMVN(np.sum(mixture_idx == key), self.mean[key], self.cov[key], S_bar, S, x_in[S])
                      for key in self.pi.keys() if np.sum(mixture_idx == key) != 0}

                rg = np.concatenate([np.concatenate(
                    [rg[key], np.tile(self.dummy[key], (np.sum(mixture_idx == key), 1))], axis=1)
                    for key in rg.keys()])

                rg_data = pd.DataFrame(rg, columns=[str(s) for s in S_bar] + [str(ca) for ca in self.cat_index])

                for val_id in S:
                    rg_data[str(val_id)] = get_given_data(val_id)

                rg_data = rg_data[sorted(rg_data.columns)]

            else:
                if len(cat_va) > 1:  # this two condition ==> grouping categorical variable
                    p = [1]
                    rg_data = gen_data_by_cat([self.revers_dummy[str(list(x[cat_va]))]], p, N)

                else:  # ==> number of categorical variable is 1

                    if x_in[cat_va] == 1:  # if dummy, y1 or y2 = 1 ==> condition on 'a' or 'b'
                        p = [1]
                        rg_data = gen_data_by_cat([self.index_cat[cat_va[0]]], p, N)

                    else:  # if y1 or y2 =0 ==> uses the complementary
                        # il faut faire le cas ou S == [] et le cas contraire !
                        cat_sample = [key for key in self.pi.keys() if key != self.index_cat[cat_va[0]]]
                        if S != []:
                            prob = [self.pi[key] * self.density_x_y_set(x, key, S) for key in cat_sample]
                            p_total = np.sum(prob)
                            prob = [p / p_total for p in prob]
                            rg_data = gen_data_by_cat(cat_sample, prob, N)
                        else:
                            prob = [self.pi[key] for key in cat_sample]
                            p_total = np.sum(prob)
                            prob = [p / p_total for p in prob]
                            rg_data = gen_data_by_cat(cat_sample, prob, N)

            y_pred = self.linear_gmm_batch(np.array(rg_data.values, dtype=np.float32))

        else:
            if len(S) == len(x_in):
                y_pred = self.linear_gmm_batch(np.expand_dims(np.array(x, dtype=np.float32), axis=0))
            else:
                S = [cu for cu in S if cu not in self.cat_index]  # get numerical index
                if cat_va == []:
                    prob = [self.pi[key] * self.density_x_y_set(x_in, key, S) for key in self.pi.keys()]
                    p_x = np.sum(prob)
                    prob = [p / p_x for p in prob]

                    mixture_idx = np.random.choice(list(self.pi.keys()), size=N, replace=True, p=prob)
                    data_a = np.tile(np.concatenate([x_in[S], np.array([1, 0])], axis=0),
                                     (np.sum(mixture_idx == 'a'), 1))
                    data_b = np.tile(np.concatenate([x_in[S], np.array([0, 1])], axis=0),
                                     (np.sum(mixture_idx == 'b'), 1))
                    data_c = np.tile(np.concatenate([x_in[S], np.array([0, 0])], axis=0),
                                     (np.sum(mixture_idx == 'c'), 1))
                    rg_data = np.concatenate([data_a, data_b, data_c], axis=0)

                    y_pred = self.linear_gmm_batch(np.array(rg_data, dtype=np.float32))
                else:
                    if len(cat_va) == 1 and x_in[cat_va] == 0:  # if dummy, y1 or y2 = 1 ==> condition on 'a' or 'b'
                        cat_sample = [key for key in self.pi.keys() if key != self.index_cat[cat_va[0]]]

                        prob = [self.pi[key] * self.density_x_y_set(x_in, key, S) for key in cat_sample]
                        p_total = np.sum(prob)
                        prob = [p / p_total for p in prob]

                        mixture_idx = np.random.choice(cat_sample, size=N, replace=True, p=prob)

                        data_1 = np.tile(np.concatenate([x_in[S], np.array(self.dummy[cat_sample[0]])], axis=0),
                                         (np.sum(mixture_idx == cat_sample[0]), 1))
                        data_2 = np.tile(np.concatenate([x_in[S], np.array(self.dummy[cat_sample[1]])], axis=0),
                                         (np.sum(mixture_idx == cat_sample[1]), 1))

                        rg_data = np.concatenate([data_1, data_2], axis=0)

                        y_pred = self.linear_gmm_batch(np.array(rg_data, dtype=np.float32))
                    else:  # on peut l'enlever grace à la première condition
                        y_pred = self.linear_gmm_batch(np.expand_dims(np.array(x_in, dtype=np.float32), axis=0))

        return np.mean(y_pred)

    def cond_exp_monte_carlo_gen(self, model, x, S, N):
        """
        Compute conditional expectation given S of the any "model" with monte carlo estimator E[f(x) | X_S = x_S]. Here, S can contains the
        index of the categorical variable.
        Args:
            model (Sklearn Ensemble): model
            x (array): observation
            S (list): index of the variable (continuous and categorical) on which on condition
            N (int): size of the monte carlo sampling

        Returns:
            float: Conditional expectation given S of any "model" with monte carlo estimator E[f(x) | X_S = x_S]
        """

        x_in = x.copy()
        cat_va = [ca for ca in S if ca in self.cat_index]  # get categorical index
        S_cu = [cu for cu in S if cu not in self.cat_index]  # get numerical index

        def get_given_data(idx):
            val = np.array([x[idx]])
            val = np.tile(val, N)
            return val

        def gen_data_by_cat(cat_va, prob, N):

            mixture_idx = np.random.choice(cat_va, size=N, replace=True, p=prob)

            rg = {key: sampleMVN(np.sum(mixture_idx == key), self.mean[key], self.cov[key], S_bar, S, x_in[S])
                  for key in cat_va if np.sum(mixture_idx == key) != 0}
            rg = np.concatenate([np.concatenate(
                [rg[key], np.tile(self.dummy[key], (np.sum(mixture_idx == key), 1))], axis=1)
                for key in rg.keys()], axis=0)

            rg_data = pd.DataFrame(rg, columns=[str(s) for s in S_bar] + [str(ca) for ca in self.cat_index])

            for val_id in S:
                rg_data[str(val_id)] = get_given_data(val_id)

            rg_data = rg_data[sorted(rg_data.columns)]
            return rg_data

        if len(S_cu) != self.d:
            cat_va = [ca for ca in S if ca in self.cat_index]  # get categorical index
            S = [cu for cu in S if cu not in self.cat_index]  # get numerical index
            S_bar = [i for i in self.index if i not in S]

            if cat_va == []:
                if S != []:
                    prob = [self.pi[key] * self.density_x_y_set(x_in, key, S) for key in self.pi.keys()]
                    p_x_y = np.sum(prob)
                    prob = [p / p_x_y for p in prob]
                    mixture_idx = np.random.choice(list(self.pi.keys()), size=N, replace=True, p=prob)
                else:
                    mixture_idx = np.random.choice(list(self.pi.keys()), size=N, replace=True, p=list(self.pi.values()))

                rg = {key: sampleMVN(np.sum(mixture_idx == key), self.mean[key], self.cov[key], S_bar, S, x_in[S])
                      for key in self.pi.keys() if np.sum(mixture_idx == key) != 0}

                # print('rg', [rg[key].shape for key in rg.keys()])
                # print('dummy', [np.tile(self.dummy[key], (np.sum(mixture_idx == key), 1)).shape for key in rg.keys()])

                rg = np.concatenate(
                    [np.concatenate([rg[key], np.tile(self.dummy[key], (np.sum(mixture_idx == key), 1))],
                                    axis=1) for key in rg.keys()], axis=0)

                rg_data = pd.DataFrame(rg, columns=[str(s) for s in S_bar] + [str(ca) for ca in self.cat_index])

                for val_id in S:
                    rg_data[str(val_id)] = get_given_data(val_id)

                rg_data = rg_data[sorted(rg_data.columns)]

            else:
                if len(cat_va) > 1:  # this two condition ==> grouping categorical variable
                    p = [1]
                    rg_data = gen_data_by_cat([self.revers_dummy[str(list(x[cat_va]))]], p, N)

                else:  # ==> number of categorical variable is 1

                    if x_in[cat_va] == 1:  # if dummy, y1 or y2 = 1 ==> condition on 'a' or 'b'
                        p = [1]
                        rg_data = gen_data_by_cat([self.index_cat[cat_va[0]]], p, N)


                    else:  # if y1 or y2 =0 ==> uses the complementary
                        # il faut faire le cas ou S == [] et le cas contraire !
                        cat_sample = [key for key in self.pi.keys() if key != self.index_cat[cat_va[0]]]
                        if S != []:
                            prob = [self.pi[key] * self.density_x_y_set(x, key, S) for key in cat_sample]
                            p_total = np.sum(prob)
                            prob = [p / p_total for p in prob]
                            rg_data = gen_data_by_cat(cat_sample, prob, N)
                        else:
                            prob = [self.pi[key] for key in cat_sample]
                            p_total = np.sum(prob)
                            prob = [p / p_total for p in prob]
                            rg_data = gen_data_by_cat(cat_sample, prob, N)

            y_pred = model.predict(np.array(rg_data.values, dtype=np.float32))

        else:
            if len(S) == len(x_in):
                y_pred = model.predict(np.expand_dims(np.array(x, dtype=np.float32), axis=0))
            else:
                S = [cu for cu in S if cu not in self.cat_index]  # get numerical index
                if cat_va == []:
                    prob = [self.pi[key] * self.density_x_y_set(x_in, key, S) for key in self.pi.keys()]
                    p_x_y = np.sum(prob)
                    prob = [p / p_x_y for p in prob]

                    mixture_idx = np.random.choice(list(self.pi.keys()), size=N, replace=True, p=prob)
                    data_a = np.tile(np.concatenate([x_in[S], np.array([1, 0])], axis=0),
                                     (np.sum(mixture_idx == 'a'), 1))
                    data_b = np.tile(np.concatenate([x_in[S], np.array([0, 1])], axis=0),
                                     (np.sum(mixture_idx == 'b'), 1))
                    data_c = np.tile(np.concatenate([x_in[S], np.array([0, 0])], axis=0),
                                     (np.sum(mixture_idx == 'c'), 1))
                    rg_data = np.concatenate([data_a, data_b, data_c], axis=0)

                    y_pred = model.predict(np.array(rg_data, dtype=np.float32))
                else:
                    if len(cat_va) == 1 and x_in[cat_va] == 0:  # if dummy, y1 or y2 = 1 ==> condition on 'a' or 'b'
                        cat_sample = [key for key in self.pi.keys() if key != self.index_cat[cat_va[0]]]

                        prob = [self.pi[key] * self.density_x_y_set(x_in, key, S) for key in cat_sample]
                        p_total = np.sum(prob)
                        prob = [p / p_total for p in prob]

                        mixture_idx = np.random.choice(cat_sample, size=N, replace=True, p=prob)

                        data_1 = np.tile(np.concatenate([x_in[S], np.array(self.dummy[cat_sample[0]])], axis=0),
                                         (np.sum(mixture_idx == cat_sample[0]), 1))
                        data_2 = np.tile(np.concatenate([x_in[S], np.array(self.dummy[cat_sample[1]])], axis=0),
                                         (np.sum(mixture_idx == cat_sample[1]), 1))

                        rg_data = np.concatenate([data_1, data_2], axis=0)

                        y_pred = model.predict(np.array(rg_data, dtype=np.float32))
                    else:
                        y_pred = model.predict(np.expand_dims(np.array(x_in, dtype=np.float32), axis=0))

        return np.mean(y_pred)

    def mar_exp_linear_mc(self, x, S):
        """
        Compute marginal expectation given S of the linear model with monte carlo estimator E[f(x_S, x_{\bar{S})]
        Args:
            x (array): observation
            S (list): index of the variable (continuous and categorical) on which on condition
            N (int): size of the sampling of monte carlo estimator

        Returns:
            float: Compute marginal expectation given S of the linear model with monte carlo estimator E[f(x_S, x_{\bar{S})]
        """
        def get_given_data(idx):
            val = np.array([x[idx]])
            val = np.tile(val, N)
            return val

        N = 50000
        if len(S) != len(x):

            rg, _ = self.gen_data(N)
            rg_data = pd.DataFrame(rg, columns=[str(n) for n in self.index] + [str(ca) for ca in self.cat_index])

            for val_id in S:
                rg_data[str(val_id)] = get_given_data(val_id)

            rg_data = rg_data[sorted(rg_data.columns)]
            y_pred = self.linear_gmm_batch(np.array(rg_data.values, dtype=np.float32))

        else:
            y_pred = self.linear_gmm_batch(np.expand_dims(np.array(x, dtype=np.float32), axis=0))

        return np.mean(y_pred)


    def cond_exp_tree(self, x, tree, S, algo="shap", data=None):
        """
        Compute conditional expectation of a tree_regressor given S: E[tree(x) | X_S = x_S]

        Args:
            x (int): observation
            tree (DecisionTreeRegressor.tree_): model
            S (list): index of the variable on which we are conditioning
            algo (string): name of the estimator. there are 5 estimators: 'pluging', 'shap', 'monte_carlo',
                            'monte_carlo_interventional', 'plugin_interventional'
            data (array): data use to compute the expectation
            N (int): size of observations sample when using Monte carlo estimators.

        Returns:
            float: E[tree(x) | X_S = x_S]
        """
        if type(data) != type(self.data):
            data = self.data

        a = tree.children_left
        b = tree.children_right
        f = tree.feature
        t = tree.threshold
        r_w = tree.weighted_n_node_samples
        r = tree.n_node_samples
        v = np.array([tree.value[i][0][0] for i in range(tree.node_count)])

        def G(i, w):

            if f[i] == -2:
                return (float(w * v[i]))
            else:
                if (f[i] in S):
                    if x[f[i]] <= t[i]:
                        return G(a[i], w)
                    else:
                        return G(b[i], w)
                else:
                    return G(a[i], w * r_w[a[i]] / r_w[i]) + G(b[i], w * r_w[b[i]] / r_w[i])

        def explore(i, tab):

            if f[i] == -2:
                tab[i] = 1
            else:
                if f[i] in S:
                    if x[f[i]] <= t[i]:
                        explore(a[i], tab)
                    else:
                        explore(b[i], tab)
                else:
                    explore(a[i], tab)
                    explore(b[i], tab)

        def get_partition(leaf_id, part):

            left = np.where(tree.children_left == leaf_id)[0]
            right = np.where(tree.children_right == leaf_id)[0]

            if (len(left) == 0) * (len(right) == 0):
                return (part)

            else:
                if len(right) != 0:
                    right = int(right[0])

                    part[f[right]] = np.concatenate((part[f[right]], np.array([[t[right], np.inf]])))
                    return get_partition(right, part)
                else:
                    left = int(left[0])
                    part[f[left]] = np.concatenate((part[f[left]], np.array([[-np.inf, t[left]]])))
                    return get_partition(left, part)

        def get_final_partition(part):
            final_partition = {}
            for i, var_part in enumerate(part):
                final_partition[i] = [np.max(var_part[:, 0]), np.min(var_part[:, 1])]
            return final_partition

        # Algorithms implemented !

        if algo == 'shap':
            return G(0, 1)

        elif algo == 'plugin':
            compatible_leaves = np.zeros(tree.node_count)
            partition_leaves = [np.array([-np.inf, np.inf]) for i in range(tree.node_count)]

            explore(0, compatible_leaves)
            compatible_leaves = (compatible_leaves == 1)
            D = np.sum(r_w[compatible_leaves])
            N = np.sum(r_w[compatible_leaves] * v[compatible_leaves])

            return N / D
        elif algo == 'plugin_v_2':
            compatible_leaves = np.zeros(tree.node_count)
            partition_leaves = [np.array([-np.inf, np.inf]) for i in range(tree.node_count)]
            compatible_leaves = np.zeros(tree.node_count)

            explore(0, compatible_leaves)

            leaf_idx = [i for i in range(tree.node_count) if compatible_leaves[i] == 1]
            if S == []:
                p = r_w[leaf_idx]/np.sum(r_w[leaf_idx])
            else:
                nbs_leaf = []
                for leaf_id in leaf_idx:
                    partition_leaves = [np.array([[-np.inf, np.inf]]) for i in range(x.shape[0])]
                    partition_leaves = get_partition(leaf_id, partition_leaves)

                    partition_leaves = pd.DataFrame(get_final_partition(partition_leaves))[S]
                    low = partition_leaves.iloc[0]
                    up = partition_leaves.iloc[1]

                    #section_x = 0
                    section_x = np.prod([(data[:, s] <= up[s])*(data[:, s] >= low[s]) for s in S], axis=0)
                    #for i in range(data.shape[0]):
                    #    section_x += np.prod([(data[i][s_bar] <= up[s_bar])*(data[i][s_bar] >= low[s_bar]) for s_bar in S_bar])
                    nbs_leaf.append(np.sum(section_x))

                nbs_leaf = np.array(nbs_leaf)

                if np.sum(nbs_leaf==0) > 0:
                    raise ValueError('division by size 0')

                p = r_w[leaf_idx]/nbs_leaf

            return np.sum(p * v[leaf_idx])

        elif algo == 'monte_carlo':
            N = 50000
            return self.cond_exp_monte_carlo_gen(tree, x, S, N)

        elif algo == 'monte_carlo_interventional':

            def get_given_data(idx):
                val = np.array([x[idx]])
                val = np.tile(val, N)
                return val

            N = 50000
            if len(S) != len(x):

                rg, _ = self.gen_data(N)
                rg_data = pd.DataFrame(rg, columns=[str(n) for n in self.index] + [str(ca) for ca in self.cat_index])

                for val_id in S:
                    rg_data[str(val_id)] = get_given_data(val_id)

                rg_data = rg_data[sorted(rg_data.columns)]
                y_pred = tree.predict(np.array(rg_data.values, dtype=np.float32))

            else:
                y_pred = tree.predict(np.expand_dims(np.array(x, dtype=np.float32), axis=0))

            return np.mean(y_pred)


        elif algo == 'plugin_interventional':
            return self.cond_exp_tree_intv(tree, x, S, v, explore, get_partition, get_final_partition, data)

        else:
            raise NotImplementedError(
                "Algorithms available are: monte_carlo, monte_carlo_interventional, plugin, shap, plugin_interventional")


    def cond_exp_tree_intv(self, model, x, S, v, explore, get_partition, get_final_partition, data):
        """ Compute marginal expectation of a tree_regressor """
        S_bar = np.array([i for i in self.index if i not in S and type(i) != str])
        x_in = x.copy()

        if len(S) != len(x_in):
            cat_va = [ca for ca in S if type(ca) == str]
            S = [cu for cu in S if type(cu) != str]

            compatible_leaves = np.zeros(model.node_count)
            explore(0, compatible_leaves)

            leaf_idx = [i for i in range(model.node_count) if compatible_leaves[i] == 1]
            leaf_proba = []
            for leaf_id in leaf_idx:
                partition_leaves = [np.array([[-np.inf, np.inf]]) for i in range(x.shape[0])]
                partition_leaves = get_partition(leaf_id, partition_leaves)

                partition_leaves = pd.DataFrame(get_final_partition(partition_leaves))[S_bar]
                low = partition_leaves.iloc[0]
                up = partition_leaves.iloc[1]

                section_x = np.sum(np.prod([(data[:, s_bar] <= up[s_bar])*(data[:, s_bar] >= low[s_bar]) for s_bar in S_bar], axis=0))
                # for i in range(data.shape[0]):
                #    section_x += np.prod([(data[i][s_bar] <= up[s_bar])*(data[i][s_bar] >= low[s_bar]) for s_bar in S_bar])

                prob_leaf = section_x/len(data)
                leaf_proba.append(prob_leaf)

            leaf_proba = np.array(leaf_proba)
            return np.sum(leaf_proba * v[compatible_leaves == 1])

        else:
            y_pred = model.predict(np.expand_dims(np.array(x, dtype=np.float32), axis=0))

        return np.mean(y_pred)

    def linear_gmm_shap(self, x, monte_carlo=False):
        """
        Compute Observational Shapley values of the linear model
        Args:
            x (array): observation
            monte_carlo (Bool): If True use monte carlo estimator otherwise exact computation
            N (int): size of the monte carlo sampling

        Returns:
            array: Observational Shapley values
        """

        m = len(x)
        va_id = list(range(m))
        if self.C[0] != []:
            for c in self.C:
                m -= len(c)
                va_id = list(set(va_id) - set(c))
            m += len(self.C)
            for c in self.C:
                va_id += [c]

        phi = np.zeros(len(x))

        for p in itertools.permutations(va_id):
            for i in range(m):
                if monte_carlo:
                    phi[chain_l(p[i])] += \
                        self.cond_exp_monte_carlo_linear(x, S=chain_l(p[:i + 1]), N=50000) \
                        - self.cond_exp_monte_carlo_linear(x, S=chain_l(p[:i]), N=50000)
                else:
                    phi[chain_l(p[i])] += \
                        self.cond_exp_linear_gmm(x, S=chain_l(p[:i + 1])) \
                        - self.cond_exp_linear_gmm(x, S=chain_l(p[:i]))
        return phi / math.factorial(m)

    def linear_gmm_mar_shap(self, x, monte_carlo=False):
        """
        Compute marginal Shapley values of the linear model
        Args:
            x (array): observation
            monte_carlo (Bool): If True use monte carlo estimator otherwise exact computation
            N (int): Size of the monte carlo sampling

        Returns:
            array: marginal shapley values
        """

        m = len(x)
        va_id = list(range(m))
        if self.C[0] != []:
            for c in self.C:
                m -= len(c)
                va_id = list(set(va_id) - set(c))
            m += len(self.C)
            for c in self.C:
                va_id += [c]

        phi = np.zeros(len(x))

        for p in itertools.permutations(va_id):
            for i in range(m):
                    phi[chain_l(p[i])] += \
                        self.mar_exp_linear_mc(x, S=chain_l(p[:i + 1])) \
                        - self.mar_exp_linear_mc(x, S=chain_l(p[:i]))
        return phi / math.factorial(m)

    def tree_shap(self, tree, x, algo, data=None):
        """
        Compute Classic Shapley values of a regressor of x

        Args:
            tree (DecisionTreeClassifier):
            x (array): observation
            algo (string): name of the estimator. there are 5 estimators: 'pluging', 'shap', 'exact', 'monte_carlo',
                            'monte_carlo_interventional', 'plugin_interventional', 'monte_carlo'
            data (array): data used to compute the Shapley values
            C (list[list]): list of the different coalition of variables by their index
            N (int): size of observations sample when using monte_carlo estimators

        Returns:
            array: Shapley values of x
        """
        if type(data) != type(self.data):
            data = self.data

        if tree is None:
            tree_model = DecisionTreeRegressor()
            tree_model = self.fit_model(tree_model).tree_
        else:
            tree_model = tree.tree_

        m = len(x)
        va_id = list(range(m))

        if self.C[0] != []:
            for c in self.C:
                m -= len(c)
                va_id = list(set(va_id) - set(c))
            m += len(self.C)
            for c in self.C:
                va_id += [c]

        phi = np.zeros(len(x))
        for p in itertools.permutations(va_id):
            for i in range(m):
                phi[chain_l(p[i])] += self.cond_exp_tree(x, tree_model, chain_l(p[:i + 1]), algo, data) - self.cond_exp_tree(
                    x, tree_model, chain_l(p[:i]), algo, data)
        return phi / math.factorial(m)


    def fit_model(self, model):
        """ Fit a model and show the errors. """
        model.fit(self.data, self.y_train)

        error_test = mean_squared_error(model.predict(self.data_test), self.y_test)
        error_train = mean_squared_error(model.predict(self.data), self.y_train)

        print('fitted model train_error = {}'.format(error_train))
        print('fitted model test_error = {}'.format(error_test))
        return model

    def shap_fit_metric(self, model, data, nb, plot=True):
        """
        Compute Shapley values of observations in data given the different estimators, and return the errors between
        the exact computation and the different estimators.

        Args:
            model (DecisionTreeRegressor):
            data (array): data used to compute the Shapley values
            nb (int): number of observations used
            plot (Bool): If True, Show a boxplot of the errors

        Returns:
            (dataFrame): data that contains the SV of each observation given the different estimators and their erros.
        """

        shap_true, shap_plugin, shap_shap, shap_monte_carlo, shap_plugin = [], [], [], [], []

        if model is None:
            tree_model = DecisionTreeRegressor()
            tree_model = self.fit_model(tree_model)
        else:
            tree_model = model

        for i in tqdm(range(nb)):
            shap_true.append(np.expand_dims(self.linear_gmm_shap(data[i]), axis=0))
            shap_plugin.append(np.expand_dims(self.tree_shap(tree_model, x=data[i], algo='plugin'), axis=0))
            shap_shap.append(np.expand_dims(self.tree_shap(tree_model, x=data[i], algo='shap'), axis=0))
            shap_monte_carlo.append(np.expand_dims(self.tree_shap(tree_model, x=data[i], algo='monte_carlo'), axis=0))
            shap_plugin.append(np.expand_dims(self.tree_shap(tree_model, x=data[i], algo='plugin_v_2'), axis=0))

        shap_true, shap_plugin, shap_shap, shap_monte_carlo, shap_plugin = np.concatenate(shap_true, axis=0), np.concatenate(
            shap_plugin, axis=0), np.concatenate(shap_shap, axis=0), np.concatenate(shap_monte_carlo, axis=0),np.concatenate(shap_plugin, axis=0)


        explainer_observational = shap.TreeExplainer(tree_model, feature_perturbation='observational')
        explainer_interventional = shap.TreeExplainer(tree_model, feature_perturbation='interventional', data=self.data)

        shap_observational = explainer_observational.shap_values(data[:nb])
        shap_interventional = explainer_interventional.shap_values(data[:nb])

        shap_data = np.concatenate([shap_true, shap_plugin, shap_shap, shap_monte_carlo, shap_plugin,
                                    shap_observational, shap_interventional], axis=0)

        shap_dataFra = pd.DataFrame(data=shap_data, columns=[str(i) for i in self.index])
        shap_dataFra['label'] = nb*['true'] + nb*['plugin'] + nb*['shap'] + nb*['mc'] +\
         nb*['plugin_sal']+ nb*['lund_obs']+ nb*['lund_int']

        err_true = l1_norm(shap_true-shap_true)
        err_plugin = l1_norm(shap_plugin-shap_true)
        err_shap = l1_norm(shap_shap-shap_true)
        err_monte_carlo = l1_norm(shap_monte_carlo-shap_true)
        err_plugin_sal = l1_norm(shap_plugin-shap_true)
        err_shap_observational = l1_norm(shap_observational-shap_true)
        err_shap_interventional = l1_norm(shap_interventional-shap_true)

        err_concat = np.concatenate([err_true, err_plugin, err_shap, err_monte_carlo, err_plugin_sal, err_shap_observational,
                                    err_shap_interventional], axis=0)

        shap_dataFra['error_l1'] = err_concat

        if plot:
            fig, ax = plt.subplots(dpi=150)
            sns.boxplot(data=shap_dataFra, x='label', y='error_l1')

        return shap_dataFra


    def cond_exp_tree_plot(self, model, data, S, nb, plot):
        """
        Compute conditional expectation of observations in data given the different estimators, and return the errors between
        the exact computation and the different estimators.

        Args:
            model (DecisionTreeRegressor):
            data (array): data used to compute the Shapley values
            S (list): index of the variable on which we condition
            nb (int): number of observations used
            plot (Bool): If True, Show a boxplot of the errors

        Returns:
            (dataFrame): data that contains the SV of each observation given the different estimators and their erros.
        """

        if model is None:
            tree_model = DecisionTreeRegressor()
            tree_model = self.fit_model(tree_model)
        else:
            tree_model = model

        e_monte_carlo, e_exact, e_shap, e_plugin, e_plugin_sal = [], [], [], [], []

        for i in tqdm(range(data[:nb].shape[0])):
            e_plugin.append(self.cond_exp_tree(x=data[i], tree=tree_model.tree_, S=S, algo="plugin"))
            e_shap.append(self.cond_exp_tree(x=data[i], tree=tree_model.tree_, S=S, algo="shap"))
            e_monte_carlo.append(self.cond_exp_tree(x=data[i], tree=tree_model.tree_, S=S, algo="monte_carlo"))
            e_exact.append(self.cond_exp_linear_gmm(x=data[i], S=S))
            e_plugin_sal.append(self.cond_exp_tree(x=data[i], tree=tree_model.tree_, S=S, algo="plugin_v_2"))

        e_monte_carlo, e_exact, e_shap, e_plugin, e_plugin_sal = \
        np.array(e_monte_carlo), np.array(e_exact), np.array(e_shap), np.array(e_plugin), np.array(e_plugin_sal)

        error_plugin = np.abs(e_plugin-e_exact)
        error_shap = np.abs(e_shap-e_exact)
        error_monte_carlo = np.abs(e_monte_carlo-e_exact)
        error_plugin_sal = np.abs(e_plugin_sal-e_exact)

        errors = np.concatenate([error_plugin, error_shap, error_monte_carlo, error_plugin_sal], axis=0)
        data_errors = pd.DataFrame(errors, columns=['errors'])
        data_errors['label'] = nb*['plugin'] + nb*['shap'] + nb*['monte_carlo'] + nb*['plugin_v_2']

        if plot:
            fig, ax = plt.subplots(dpi=200)
            sns.boxplot(data=data_errors, x='label', y='errors')

            # ax.plot(range(len(e_monte_carlo)), e_monte_carlo, label="monte_carlo")
            # ax.plot(range(len(e_plugin)), e_plugin, label="plugin")
            # ax.plot(range(len(e_shap)), e_shap, label="shap")
            # ax.plot(range(len(e_exact)), e_exact, label="exact")
            # ax.legend()

            # plt.figure(dpi=200)
            # plt.bar(['mae_plugin', 'mae_shap', 'mar_monte_carlo'], errors, width=0.4)
            # plt.legend()

            # print('nic = {}, algo = {}, monte_carlo = {}'.format(error_plugin, error_shap, error_monte_carlo))

        return {'err_plugin': error_plugin, 'err_shap': error_shap, 'err_monte_carlo': error_monte_carlo,
                'err_plugin_sal':error_plugin_sal},\
                {'e_plugin': e_plugin, 'e_shap': e_shap, 'e_monte_carlo': e_monte_carlo,
                        'e_plugin_sal':e_plugin_sal}
